: # use perl
eval 'exec $ANTELOPE/bin/perl -S $0 "$@"'
if 0;

use lib ( "$ENV{ANTELOPE}/data/perl", "/opt/antelope/dev-64/src/bin/rt/dbevents" ) ;

#   Copyright (c) 2010 Boulder Real Time Technologies, Inc.           
#                                                                     
#   This software module is wholly owned by Boulder Real Time         
#   Technologies, Inc. Any use of this software module without        
#   express written permission from Boulder Real Time Technologies,   
#   Inc. is prohibited.             

$Tk::EventSpreadsheet::VERSION = '1.0' ;

package Tk::EventSpreadsheet ;

use Tk::widgets qw(Canvas Button Btable/Bspreadsheet) ;
use base qw(Tk::Frame) ;

use strict ;
use warnings ;
use Datascope ;

use DbEventView ;

sub trim {
	my $string = shift ;

	$string =~ s/^\s+// ;
	$string =~ s/\s+$// ;

	return $string ;
}

sub process_netmag {
	my @db ;
	$db[0] = shift ;
	$db[1] = shift ;
	$db[2] = shift ;
	$db[3] = shift ;
	my $magtype = shift ;

	my $table_name = dbquery ( @db, "dbTABLE_NAME" ) ;

	if ($table_name eq "origin") {
		my @dbn = dblookup ( @db, 0, 'netmag', 0, 0 ) ;
		my @rows = dbmatches ( @db, @dbn, "netmag_hook", "orid#orid" ) ;
		@db = @dbn ;
		foreach my $row ( @rows ) {
			$db[3] = $row ;
			if ( $magtype eq dbgetv ( @db, "magtype" ) ) {
				return @db ;
			}
		}
	}

	return ;
}

sub print_netmag_latency {
	my @db ;
	$db[0] = shift ;
	$db[1] = shift ;
	$db[2] = shift ;
	$db[3] = shift ;
	my $magtype = shift ;

	my $table_name = dbquery ( @db, "dbTABLE_NAME" ) ;

	if ($table_name eq "origin") {
		my ($otime, $olddate) = dbgetv ( @db, "time", "lddate" ) ;
		my $oval = trim epoch2str ( $olddate-$otime, "%H:%M:%S" ) ;
		my @dbn = dblookup ( @db, 0, 'netmag', 0, 0 ) ;
		my @rows = dbmatches ( @db, @dbn, "netmag_hook", "orid#orid" ) ;
		@db = @dbn ;
		foreach my $row ( @rows ) {
			$db[3] = $row ;
			if ( $magtype eq dbgetv ( @db, "magtype" ) ) {
				my $lddate = dbgetv ( @db, "lddate" ) ;
				my $val = trim epoch2str ( $lddate-$otime, "%H:%M:%S" ) ;
				if ( $val eq $oval) { return '        ' ;}
				return $val ; 
			}
		}
	}

	return '        ' ;
}

sub printmag {
	my @db ;
	$db[0] = shift ;
	$db[1] = shift ;
	$db[2] = shift ;
	$db[3] = shift ;
	my $magtype = shift ;

	my $magnitude ;

	my @dbn = process_netmag ( @db, $magtype ) ;

	if ( ! @dbn ) {
		if ($magtype eq "ml") {
			$magnitude = dbgetv (@db, "ml");
		} elsif ($magtype eq "mb") {
			$magnitude = dbgetv (@db, "mb");
		} elsif ($magtype eq "ms") {
			$magnitude = dbgetv (@db, "ms");
		}
	} else {
		$magnitude = dbgetv (@dbn, "magnitude");
	}

	if ( ! defined $magnitude ) {return "   " ; }
	if ($magnitude == -999.0) {$magnitude = "   " ;}
	else {$magnitude = sprintf ("%.1f", $magnitude);}

	return $magnitude ;
}

my %col_events_default = (
	evid => {
		label => 'evid',
		print_callback => 'evid',
		description => 'event id',
		anchor => 'e',
		width_template => ' XXXXXX ',
		sort_type => 'numerical',
	},
	orid => {
		label => 'orid',
		description => 'preferred origin id',
		print_callback => 'orid',
		anchor => 'e',
		width_template => ' XXXXXX ',
		sort_type => 'numerical',
	},
	time => {
		label => 'origin time',
		description => 'preferred origin time',
		print_callback => [ sub {
				my @db ;
				$db[0] = shift ;
				$db[1] = shift ;
				$db[2] = shift ;
				$db[3] = shift ;
				my $time = dbgetv ( @db, 'time' ) ;
				my $val = trim epoch2str ( $time, "%y%j %m/%d %T" ) ;
				return $val ;
			} ] ,
		anchor => 'e',
		width_template => ' XXXXX XXXXX XXXXXXXXXXXX ',
	},
	author => {
		label => 'auth',
		description => 'preferred origin author',
		print_callback => 'auth',
		anchor => 'e',
		width_template => ' XXXX   ',
	},
	review => {
		label => 'rev',
		description => 'preferred origin reviewed status - * means automatic magnitudes have been finalized',
		print_callback => [ sub {
				my @db ;
				$db[0] = shift ;
				$db[1] = shift ;
				$db[2] = shift ;
				$db[3] = shift ;
				my $review = dbgetv ( @db, 'review' ) ;
				if ($review eq "-") {
					return " " ;
				} elsif ($review eq "M") {
					return "*" ;
				} else {
					return $review ;
				}
			} ] ,
		anchor => 'e',
		width_template => ' XXX   ',
	},
	depth => {
		label => 'depth',
		description => 'preferred origin depth',
		print_callback => [ sub {
				my @db ;
				$db[0] = shift ;
				$db[1] = shift ;
				$db[2] = shift ;
				$db[3] = shift ;
				my $depth = dbgetv ( @db, 'depth' ) ;
				return sprintf ( "%.2f", $depth ) ;
			} ] ,
		anchor => 'e',
		width_template => ' XXXXX   ',
		sort_type => 'numerical',
	},
	mb => {
		label => 'mb',
		description => 'preferred origin mb magnitude',
		print_callback => [ \&printmag, "mb" ],
		anchor => 'e',
		width_template => ' XXX   ',
		sort_type => 'numerical',
	},
	ms => {
		description => 'preferred origin ms magnitude',
		label => 'ms',
		print_callback => [ \&printmag, "ms" ],
		anchor => 'e',
		width_template => ' XXX   ',
		sort_type => 'numerical',
	},
	mwp => {
		label => 'mwp',
		description => 'preferred origin mwp magnitude',
		print_callback => [ \&printmag, "mwp" ],
		anchor => 'e',
		width_template => '  XXX   ',
		sort_type => 'numerical',
	},
	mw => {
		label => 'mw',
		description => 'preferred origin mw magnitude',
		print_callback => [ \&printmag, "mw" ],
		anchor => 'e',
		width_template => ' XXX   ',
		sort_type => 'numerical',
	},
	ndef => {
		label => 'ndef',
		description => 'preferred origin number of defining phases',
		print_callback => 'ndef',
		anchor => 'e',
		width_template => ' XXX   ',
		sort_type => 'numerical',
	},
	nass => {
		label => 'nass',
		description => 'preferred origin number of associated phases',
		print_callback => 'nass',
		anchor => 'e',
		width_template => ' XXX   ',
		sort_type => 'numerical',
	},	
	latency => {
		label => 'latency',
		description => 'preferred origin latency between time origin added to database and origin time',
		print_callback => [ sub {
				my @db ;
				$db[0] = shift ;
				$db[1] = shift ;
				$db[2] = shift ;
				$db[3] = shift ;
				my ($time, $lddate) = dbgetv ( @db, 'time', 'lddate' ) ;
				my $val = trim epoch2str ( $lddate-$time, "%H:%M:%S" ) ;
				return $val ; 
			} ],
		anchor => 'e',
		width_template => ' XXXXXXXX ',
	},
	region => {
		label => 'region',
		description => 'preferred origin region',
		print_callback => 'grname(lat,lon)',
		anchor => 'w',
		width_template => ' XXXXXXXXXXXXXXXXXXXXXXXXXXX ',
	},
) ;

my @coldisplay_events_default = (
	'evid',
	'orid',
	'time',
	'author',
	'review',
	'depth',
	'mb',
	'ms',
	'mwp',
	'mw',
	'ndef',
	'nass',
	'latency',
	'region',
) ;

my %col_event_default = (
	evid => {
		label => 'evid',
		print_callback => 'evid',
		description => 'event id',
		anchor => 'e',
		width_template => ' XXXXXX ',
		sort_type => 'numerical',
	},
	orid => {
		label => 'orid',
		description => 'origin id',
		print_callback => 'orid',
		anchor => 'e',
		width_template => ' XXXXXX ',
		sort_type => 'numerical',
	},
	time => {
		label => 'origin time',
		description => 'origin time',
		print_callback => [ sub {
				my @db ;
				$db[0] = shift ;
				$db[1] = shift ;
				$db[2] = shift ;
				$db[3] = shift ;
				my $time = dbgetv ( @db, 'time' ) ;
				my $val = trim epoch2str ( $time, "%y%j %m/%d %T" ) ;
				return $val ;
			} ] ,
		anchor => 'e',
		width_template => ' XXXXX XXXXX XXXXXXXXXXXX ',
	},
	author => {
		label => 'auth',
		description => 'author',
		print_callback => 'auth',
		anchor => 'e',
		width_template => ' XXXX   ',
	},
	review => {
		label => 'rev',
		description => 'reviewed status - * means automatic magnitudes have been finalized',
		print_callback => [ sub {
				my @db ;
				$db[0] = shift ;
				$db[1] = shift ;
				$db[2] = shift ;
				$db[3] = shift ;
				my $review = dbgetv ( @db, 'review' ) ;
				if ($review eq "-") {
					return " " ;
				} elsif ($review eq "M") {
					return "*" ;
				} else {
					return $review ;
				}
			} ] ,
		anchor => 'e',
		width_template => ' XXX   ',
	},
	lat => {
		label => 'latitude',
		description => 'latitude',
		print_callback => [ sub {
				my @db ;
				$db[0] = shift ;
				$db[1] = shift ;
				$db[2] = shift ;
				$db[3] = shift ;
				my $lat = dbgetv ( @db, 'lat' ) ;
				return sprintf ( "%.4f", $lat ) ;
			} ] ,
		anchor => 'e',
		width_template => ' XXXXXXXX   ',
		sort_type => 'numerical',
	},
	lon => {
		label => 'longitude',
		description => 'longitude',
		print_callback => [ sub {
				my @db ;
				$db[0] = shift ;
				$db[1] = shift ;
				$db[2] = shift ;
				$db[3] = shift ;
				my $lon = dbgetv ( @db, 'lon' ) ;
				return sprintf ( "%.4f", $lon ) ;
			} ] ,
		anchor => 'e',
		width_template => ' XXXXXXXXX   ',
		sort_type => 'numerical',
	},
	depth => {
		label => 'depth',
		description => 'depth',
		print_callback => [ sub {
				my @db ;
				$db[0] = shift ;
				$db[1] = shift ;
				$db[2] = shift ;
				$db[3] = shift ;
				my $depth = dbgetv ( @db, 'depth' ) ;
				return sprintf ( "%.2f", $depth ) ;
			} ] ,
		anchor => 'e',
		width_template => ' XXXXX   ',
		sort_type => 'numerical',
	},
	mb => {
		label => 'mb',
		description => 'mb magnitude',
		print_callback => [ \&printmag, "mb" ],
		anchor => 'e',
		width_template => ' XXX   ',
		sort_type => 'numerical',
	},
	mblatency => {
		label => 'mb ltncy',
		description => 'latency between time mb added to database and origin time',
		print_callback => [ \&print_netmag_latency, "mb" ],
		anchor => 'c',
		width_template => ' XXXXXXXX ',
	},
	ms => {
		description => 'ms magnitude',
		label => 'ms',
		print_callback => [ \&printmag, "ms" ],
		anchor => 'e',
		width_template => ' XXX   ',
		sort_type => 'numerical',
	},
	mslatency => {
		label => 'ms ltncy',
		description => 'latency between time ms added to database and origin time',
		print_callback => [ \&print_netmag_latency, "ms" ],
		anchor => 'c',
		width_template => ' XXXXXXXX ',
	},
	mwp => {
		label => 'mwp',
		description => 'mwp magnitude',
		print_callback => [ \&printmag, "mwp" ],
		anchor => 'e',
		width_template => '  XXX   ',
		sort_type => 'numerical',
	},
	mwplatency => {
		label => 'mwp ltncy',
		description => 'latency between time mwp added to database and origin time',
		print_callback => [ \&print_netmag_latency, "mwp" ],
		anchor => 'c',
		width_template => '  XXXXXXXX ',
	},
	mw => {
		label => 'mw',
		description => 'mw magnitude',
		print_callback => [ \&printmag, "mw" ],
		anchor => 'e',
		width_template => ' XXX   ',
		sort_type => 'numerical',
	},
	ndef => {
		label => 'ndef',
		description => 'number of defining phases',
		print_callback => 'ndef',
		anchor => 'e',
		width_template => ' XXX   ',
		sort_type => 'numerical',
	},
	nass => {
		label => 'nass',
		description => 'number of associated phases',
		print_callback => 'nass',
		anchor => 'e',
		width_template => ' XXX   ',
		sort_type => 'numerical',
	},	
	latency => {
		label => 'latency',
		description => 'latency between time origin added to database and origin time',
		print_callback => [ sub {
				my @db ;
				$db[0] = shift ;
				$db[1] = shift ;
				$db[2] = shift ;
				$db[3] = shift ;
				my ($time, $lddate) = dbgetv ( @db, 'time', 'lddate' ) ;
				my $val = trim epoch2str ( $lddate-$time, "%H:%M:%S" ) ;
				return $val ; 
			} ],
		anchor => 'w',
		width_template => ' XXXXXXXX ',
	},
	region => {
		label => 'region',
		description => 'region',
		print_callback => 'grname(lat,lon)',
		anchor => 'w',
		width_template => ' XXXXXXXXXXXXXXXXXXXXXXXXXXX ',
	},
) ;

my @coldisplay_event_default = (
	'orid',
	'time',
	'author',
	'review',
	'lat',
	'lon',
	'depth',
	'mb',
	'mblatency',
	'ms',
	'mwp',
	'mwplatency',
	'mw',
	'ndef',
	'nass',
	'latency',
) ;

Construct Tk::Widget 'EventSpreadsheet' ;

sub ClassInit {
	my ($class, $mw) = @_ ;
	$class->SUPER::ClassInit($mw) ;

	my $bitsup = pack ( 
		"b8" x 5,
		"........",
		"...11...",
		"..1111..",
		".111111.",
		"........"
	) ;
	$mw->DefineBitmap ( 'arrowup' => 8, 5, $bitsup ) ;
	my $bitsdown = pack ( 
		"b8" x 5,
		"........",
		".111111.",
		"..1111..",
		"...11...",
		"........"
	) ;
	$mw->DefineBitmap ( 'arrowdown' => 8, 5, $bitsdown ) ;
	my $bitsnone = pack ( 
		"b8" x 5,
		"........",
		"........",
		"........",
		"........",
		"........"
	) ;
	$mw->DefineBitmap ( 'arrownone' => 8, 5, $bitsnone ) ;
}

sub Populate {
	my $self = shift ;
	my ($args) = @_ ;

	$self->ConfigSpecs (
		'DEFAULT'		=> $self,
		'-db'			=> ['METHOD', 'db', 'Db', undef],
		'-col_events'		=> ['PASSIVE', 'colevents', 'ColEvents', 
						\%col_events_default],
		'-coldisplay_events'	=> ['METHOD', 'coldisplayevents', 'ColDisplayEvents', 
						\@coldisplay_events_default],
		'-sort_events'		=> ['METHOD', 'sortevents', 'SortEvents', 
						'time'],
		'-sort_events_reversed'	=> ['METHOD', 'sorteventsreversed', 'SortEventsReversed', 
						1],
		'-col_event'		=> ['PASSIVE', 'colevent', 'ColEvent', 
						\%col_event_default],
		'-coldisplay_event'	=> ['METHOD', 'coldisplayevent', 'ColDisplayEvent', 
						\@coldisplay_event_default],
		'-sort_event'		=> ['METHOD', 'sortevent', 'SortEvent', 
						'latency'],
		'-sort_event_reversed'	=> ['METHOD', 'sorteventreversed', 'SortEventReversed', 
						1],
		'-update_callback'	=> ['METHOD', 'updatecallback', 'UpdateCallback', 
						undef],
		'-event_selection'	=> ['METHOD', 'eventselection', 'EventSelection', 
						undef],
		) ;

	$self->SUPER::Populate($args) ;

	my $scrollbar_events = $self->Scrollbar (
		-orient => 'vertical',
		-width => 10,
		)->grid (
			-row => 0,
			-column => 0,
			-sticky => 'nse'
			) ;

	my $events_table = $self->Component ( 'Bspreadsheet' => 'events_table' ) ;

	$events_table->configure (
	 	-width => 0,
	 	-height => 51,
		-colstretchmode => 'last',
		-titlerows => 1,
		-titlecols => 1,
		-background => 'white',
		-font => ['courier', 10, 'normal'],
		-maxwidth => 2000,
		-maxheight => 1000,
		-state => 'disabled',
		-selectmode => 'browse',
		-selecttype => 'cell',
		-selecttitle => 1,
		-cursor => 'cross',
		-yscrollcommand => ['set' => $scrollbar_events],
		-command => [ sub {
				my $setflag = shift ;
				my $row = shift ;
				my $col = shift ;
				&cellcmd_events ( $self, $events_table, $row, $col ) ;
			} ],
	 	) ;

	$events_table-> grid (
		-row => 0,
		-column => 1,
		-sticky => 'nsew',
		) ;

	$scrollbar_events->configure (
		-command => ['yview' => $events_table],
		) ;
	$self->gridRowconfigure ( 0, -weight => 0 ) ;
	$self->gridColumnconfigure ( 0, -weight => 0 ) ;
	$self->gridColumnconfigure ( 1, -weight => 1 ) ;

	$events_table->tagConfigure ( 'events1',
		-background => 'white',
		) ;
	$events_table->tagConfigure ( 'events2',
		-background => "#e0e0ff",
		) ;
	$events_table->tagConfigure ( 'title',
		-background => "#fff0e0",
		) ;
	$events_table->tagConfigure ( 'selectedevents',
		-foreground => "black",
		-background => "#ffe0e0",
		) ;

	$events_table->bindtags ( undef ) ;

	$events_table->bind ( '<ButtonPress-1>', [ sub {
					my $W = shift ;
					my $e = $W->XEvent ;
					my $x = $e->x ;
					my $y = $e->y ;
					my $index = '@' . $x . "," . $y ;
					my ($row, $col) = split ",", $W->index($index) ;
					if ($row == 0) {return;}
					$self->set_event_selection_from_table ( $events_table, $row-1 ) ;
					my $event_table = $self->Subwidget ( 'event_table' ) ;
					$self->set_origin_selection ( $event_table, 'prefor' ) ;
				}, ] ) ;


	my $scrollbar_event = $self->Scrollbar (
		-orient => 'vertical',
		-width => 10,
		)->grid (
			-row => 1,
			-column => 0,
			-sticky => 'nse'
			) ;

	my $event_table = $self->Component ( 'Bspreadsheet' => 'event_table' ) ;

	$event_table->configure (
	 	-width => 0,
	 	-height => 11,
		-colstretchmode => 'last',
		-titlerows => 1,
		-titlecols => 1,
		-background => 'white',
		-font => ['courier', 10, 'normal'],
		-maxwidth => 2000,
		-maxheight => 1000,
		-state => 'disabled',
		-selectmode => 'browse',
		-selecttype => 'cell',
		-selecttitle => 1,
		-cursor => 'cross',
		-yscrollcommand => ['set' => $scrollbar_event],
		-command => [ sub {
				my $setflag = shift ;
				my $row = shift ;
				my $col = shift ;
				&cellcmd_event ( $self, $event_table, $row, $col ) ;
			} ],
	 	) ;

	$event_table-> grid (
		-row => 1,
		-column => 1,
		-sticky => 'nsew',
		) ;

	$scrollbar_event->configure (
		-command => ['yview' => $event_table],
		) ;
	$self->gridRowconfigure ( 1, -weight => 1 ) ;
	$self->gridColumnconfigure ( 0, -weight => 0 ) ;
	$self->gridColumnconfigure ( 1, -weight => 1 ) ;

	$event_table->tagConfigure ( 'origins1',
		-background => "#d0ffd0",
		) ;
	$event_table->tagConfigure ( 'origins2',
		-background => "white",
		) ;
	$event_table->tagConfigure ( 'title',
		-background => "#fff0ff",
		) ;
	$event_table->tagConfigure ( 'porigins1',
		-foreground => "red",
		-background => "#d0ffd0",
		) ;
	$event_table->tagConfigure ( 'porigins2',
		-foreground => "#d00000",
		-background => "white",
		) ;
	$event_table->tagConfigure ( 'selectedorigins',
		-foreground => "black",
		-background => "#ffe0e0",
		) ;
	$event_table->tagConfigure ( 'pselectedorigins',
		-foreground => "#d00000",
		-background => "#ffe0e0",
		) ;

	$event_table->bindtags ( undef ) ;

	$event_table->bind ( '<ButtonPress-1>', [ sub {
					my $W = shift ;
					my $e = $W->XEvent ;
					my $x = $e->x ;
					my $y = $e->y ;
					my $index = '@' . $x . "," . $y ;
					my ($row, $col) = split ",", $W->index($index) ;
					if ($row == 0) {return;}
					$self->set_origin_selection_from_table ( $event_table, $row-1 ) ;
				}, ] ) ;

	$self->{ess_params}{colwidths_events} = [ ] ;
	$self->{ess_params}{col_events_buttons} = { } ;
	$self->{ess_params}{col_event_buttons} = { } ;

	$self->{ess_params}{colwidths_event} = [ ] ;
}

sub set_origin_selection {
	my $self = shift ;
	my $table = shift ;
	my $selected_dbrow = shift ;

	if ($selected_dbrow eq "prefor") {
		$self->{ess_params}{evv}->put ( -selected_origin => 'prefor' ) ;
	} else {
		if ( $selected_dbrow < 0 ) { return; }

		my @db = @{$self->{ess_params}{db_ref}} ;
		my @dbo = dblookup ( @db, 0, 'origin', 0, 0 ) ;
		$dbo[3] = $selected_dbrow ;
		my $orid = dbgetv ( @dbo, "orid" ) ;
		$self->{ess_params}{evv}->put ( -selected_origin => $orid ) ;
	}
	$self->{ess_params}{evv}->update_selected_origin ;
	($self->{ess_params}{selected_origin_dbrow}) = $self->{ess_params}{evv}->get ( 'selected_origin_dbrow' ) ;
	($self->{ess_params}{selected_origin_sortedrow}) = $self->{ess_params}{evv}->get ( 'selected_origin_sortedrow' ) ;

	my $start = $self->{ess_params}{selected_origin_sortedrow}+1 . ",0" ;
	# print "origin at $start\n" ;
	$table->see( $start ) ;

	$table->configure (
		-state => 'normal',
		) ;
	$table->update ;
}

sub event_selection {
	my $self = shift ;
	my $selected_evid = shift ;

	if ( ! defined $selected_evid ) {
		return $self->{ess_params}{evv}->get ( -selected_event ) ;
	}

	if ( $selected_evid < 0 ) { return; }

	$self->{ess_params}{evv}->put ( -selected_event => $selected_evid ) ;
	$self->{ess_params}{evv}->update_selected_event ;
	($self->{ess_params}{selected_event_dbrow}) = $self->{ess_params}{evv}->get ( 'selected_event_dbrow' ) ;
	($self->{ess_params}{selected_event_sortedrow}) = $self->{ess_params}{evv}->get ( 'selected_event_sortedrow' ) ;
	($self->{ess_params}{selected_origin_dbrow}) = $self->{ess_params}{evv}->get ( 'selected_origin_dbrow' ) ;
	($self->{ess_params}{selected_origin_sortedrow}) = $self->{ess_params}{evv}->get ( 'selected_origin_sortedrow' ) ;
	($self->{ess_params}{ssrow2dbrow_selected}) = $self->{ess_params}{evv}->get ( 'selected_sortedrow2dbrow' ) ;
	($self->{ess_params}{dbrow2ssrow_selected}) = $self->{ess_params}{evv}->get ( 'selected_dbrow2sortedrow' ) ;

	my $start = $self->{ess_params}{dbrow2ssrow_events}->[$self->{ess_params}{selected_event_dbrow}]+1 . ",0" ;
	# print "event at $start\n" ;
	my $dbt = $self->Subwidget ( 'events_table' ) ;
	$dbt->see( $start ) ;

	$dbt->configure (
		-state => 'normal',
		) ;
	$dbt->update ;

	$dbt = $self->Subwidget ( 'event_table' ) ;
	my $nrows = @{$self->{ess_params}{ssrow2dbrow_selected}} ;
	$dbt->configure (
		-rows	=> $nrows + 1,
		) ;
}

sub set_event_selection {
	my $self = shift ;
	my $table = shift ;
	my $selected_dbrow = shift ;

	if ( $selected_dbrow < 0 ) { return; }

	my @db = @{$self->{ess_params}{db_ref}} ;
	my @dbo = dblookup ( @db, 0, 'origin', 0, 0 ) ;
	$dbo[3] = $selected_dbrow ;
	my $evid = dbgetv ( @dbo, "evid" ) ;
	$self->{ess_params}{evv}->put ( -selected_event => $evid ) ;
	$self->{ess_params}{evv}->update_selected_event ;
	($self->{ess_params}{selected_event_dbrow}) = $self->{ess_params}{evv}->get ( 'selected_event_dbrow' ) ;
	($self->{ess_params}{selected_event_sortedrow}) = $self->{ess_params}{evv}->get ( 'selected_event_sortedrow' ) ;
	($self->{ess_params}{selected_origin_dbrow}) = $self->{ess_params}{evv}->get ( 'selected_origin_dbrow' ) ;
	($self->{ess_params}{selected_origin_sortedrow}) = $self->{ess_params}{evv}->get ( 'selected_origin_sortedrow' ) ;
	($self->{ess_params}{ssrow2dbrow_selected}) = $self->{ess_params}{evv}->get ( 'selected_sortedrow2dbrow' ) ;
	($self->{ess_params}{dbrow2ssrow_selected}) = $self->{ess_params}{evv}->get ( 'selected_dbrow2sortedrow' ) ;

	my $start = $self->{ess_params}{dbrow2ssrow_events}->[$selected_dbrow]+1 . ",0" ;
	# print "event at $start\n" ;
	$table->see( $start ) ;

	$table->configure (
		-state => 'normal',
		) ;
	$table->update ;

	my $dbt = $self->Subwidget ( 'event_table' ) ;
	my $nrows = @{$self->{ess_params}{ssrow2dbrow_selected}} ;
	$dbt->configure (
		-rows	=> $nrows + 1,
		) ;
}

sub set_event_selection_from_table {
	my $self = shift ;
	my $table = shift ;
	my $row = shift ;

	my $selected_dbrow = $self->{ess_params}{ssrow2dbrow_events}->[$row] ;

	set_event_selection ( $self, $table, $selected_dbrow ) ;
}

sub set_origin_selection_from_table {
	my $self = shift ;
	my $table = shift ;
	my $row = shift ;

	my $selected_dbrow = $self->{ess_params}{ssrow2dbrow_selected}->[$row] ;

	set_origin_selection ( $self, $table, $selected_dbrow ) ;
}

sub db {
	my $self = shift ;
	my $value = shift ;

	if ( ! defined $value ) {
		$value = $self->{ess_params}{db_ref} ;
		return $value ;
	}
	$self->{ess_params}{db_ref} = $value ;

	# print "In db\n" ;

	my @db = @{$value} ;

	$self->{ess_params}{evv} = DbEventView->new ( $self, @db ) ;

	$self->{ess_params}{evv}->update ;

	$self->{ess_params}{evv}->put ( -update_callback => [ sub {
				my $evv = shift ;
				my $ess = shift ;
				my $events_table = $ess->Subwidget ( 'events_table' ) ;
				my $event_table = $ess->Subwidget ( 'event_table' ) ;
				$events_table->configure (
					-state => 'normal',
					) ;
				$events_table->update ;
				$event_table->configure (
					-state => 'normal',
					) ;
				$event_table->update ; }, $self ] ) ;

	($self->{ess_params}{ssrow2dbrow_events}) = $self->{ess_params}{evv}->get ( 'sortedrow2dbrow' ) ;
	($self->{ess_params}{dbrow2ssrow_events}) = $self->{ess_params}{evv}->get ( 'dbrow2sortedrow' ) ;
	($self->{ess_params}{selected_event_dbrow}) = $self->{ess_params}{evv}->get ( 'selected_event_dbrow' ) ;
	($self->{ess_params}{ssrow2dbrow_selected}) = $self->{ess_params}{evv}->get ( 'selected_sortedrow2dbrow' ) ;
	($self->{ess_params}{dbrow2ssrow_selected}) = $self->{ess_params}{evv}->get ( 'selected_dbrow2sortedrow' ) ;
	($self->{ess_params}{selected_origin_dbrow}) = $self->{ess_params}{evv}->get ( 'selected_origin_dbrow' ) ;

	if ( defined $self->{ess_params}{sort_events_need} ) {
		# print "In db calling sort_events\n" ;
		$self->configure ( -sort_events => $self->{ess_params}{sort_events} ) ;
	}

	if ( defined $self->{ess_params}{coldisplay_events_need} ) {
		# print "In db calling coldisplay_events\n" ;
		$self->configure ( -coldisplay_events => $self->{ess_params}{coldisplay_events} ) ;
	}

	if ( defined $self->{ess_params}{sort_event_need} ) {
		# print "In db calling sort_event\n" ;
		$self->configure ( -sort_event => $self->{ess_params}{sort_event} ) ;
	}

	if ( defined $self->{ess_params}{coldisplay_event_need} ) {
		# print "In db calling coldisplay_event\n" ;
		$self->configure ( -coldisplay_event => $self->{ess_params}{coldisplay_event} ) ;
	}

	if ( defined $self->{ess_params}{update_callback_need} ) {
		# print "In db calling update_callback\n" ;
		$self->configure ( -update_callback => $self->{ess_params}{update_callback} ) ;
	}
	# print "Leaving db\n" ;
}

sub update_callback {
	my $self = shift ;
	my $value = shift ;

	if ( ! defined $value ) {
		$value = $self->{ess_params}{update_callback} ;
		return $value ;
	}
	$self->{ess_params}{update_callback} = $value ;

	if ( ! defined $self->{ess_params}{evv} ) {
		# print "In update_callback with undef evv\n" ;
		$self->{ess_params}{update_callback_need} = 1 ;
		return ;
	}
	undef $self->{ess_params}{update_callback_need} ;
	# print "In update_callback with defined evv\n" ;

	$self->{ess_params}{evv}->put ( -update_event_callback => $self->{ess_params}{update_callback} ) ;
}

sub sort_events_private {
	my $self = shift ;

	my $name = $self->{ess_params}{sort_events} ;
	my $reversed = $self->{ess_params}{sort_events_reversed} ;
	my $colarr_ref = $self->cget(-col_events) ;

	if (defined $name) {
		if ( defined $colarr_ref->{$name}{sort_callback} ) {
			$self->{ess_params}{evv}->put (
				-sort_events_value_callback => $colarr_ref->{$name}{sort_callback},
				) ;
		} elsif ( defined $colarr_ref->{$name}{print_callback} ) {
			$self->{ess_params}{evv}->put (
				-sort_events_value_callback => $colarr_ref->{$name}{print_callback},
				) ;
		}
		if ( defined $colarr_ref->{$name}{sort_type} ) {
			$self->{ess_params}{evv}->put (
				-sort_events_type => $colarr_ref->{$name}{sort_type},
				) ;
		} else {
			$self->{ess_params}{evv}->put (
				-sort_events_type => 'string',
				) ;
		}
	}

	if (defined $reversed) {
		$self->{ess_params}{evv}->put (
			-sort_events_reversed => $reversed,
			) ;
	}

	if (defined $name || defined $reversed) {
		$self->{ess_params}{evv}->update_sort ;

		($self->{ess_params}{ssrow2dbrow_events}) = $self->{ess_params}{evv}->get ( 'sortedrow2dbrow' ) ;
		($self->{ess_params}{dbrow2ssrow_events}) = $self->{ess_params}{evv}->get ( 'dbrow2sortedrow' ) ;
		($self->{ess_params}{selected_event_dbrow}) = $self->{ess_params}{evv}->get ( 'selected_event_dbrow' ) ;

		my $dbt = $self->Subwidget ( 'events_table' ) ;
		my $selected_dbrow = $self->{ess_params}{selected_event_dbrow} ;

		set_event_selection ( $self, $dbt, $selected_dbrow ) ;

		$dbt->configure (
			-state => 'normal',
			) ;
		$dbt->update ;

		$dbt = $self->Subwidget ( 'event_table' ) ;
		$dbt->configure (
			-state => 'normal',
			) ;
		$dbt->update ;
	}
}

sub sort_events {
	my $self = shift ;
	my $value = shift ;

	if ( ! defined $value ) {
		$value = $self->{ess_params}{sort_events} ;
		return $value ;
	}

	$self->{ess_params}{sort_events} = $value ;

	if ( ! defined $self->{ess_params}{evv} ) {
		# print "In sort_events with undef evv\n" ;
		$self->{ess_params}{sort_events_need} = 1 ;
		return ;
	}
	undef $self->{ess_params}{sort_events_need} ;
	# print "In sort_events sorting on $value\n" ;

	sort_events_private ( $self ) ;

	&setup_title_buttons ( $self, $self->cget(-coldisplay_events), $self->cget(-col_events), 
				\$self->{ess_params}{sort_events},
				\$self->{ess_params}{sort_events_reversed},
				$self->{ess_params}{col_events_buttons},
				\&sort_events_private ) ;
}

sub sort_events_reversed {
	my $self = shift ;
	my $value = shift ;

	if ( ! defined $value ) {
		$value = $self->{ess_params}{sort_events_reversed} ;
		return $value ;
	}

	$self->{ess_params}{sort_events_reversed} = $value ;

	if ( ! defined $self->{ess_params}{evv} ) {
		# print "In sort_events_reversed with undef evv\n" ;
		$self->{ess_params}{sort_events_reversed_need} = 1 ;
		return ;
	}
	undef $self->{ess_params}{sort_events_reversed_need} ;
	# print "In sort_events_reversed sorting on $value\n" ;

	sort_events_private ( $self ) ;
}

sub sort_event_private {
	my $self = shift ;

	my $name = $self->{ess_params}{sort_event} ;
	my $reversed = $self->{ess_params}{sort_event_reversed} ;
	my $colarr_ref = $self->cget(-col_event) ;

	if (defined $name) {
		if ( defined $colarr_ref->{$name}{sort_callback} ) {
			$self->{ess_params}{evv}->put (
				-sort_selected_event_value_callback => $colarr_ref->{$name}{sort_callback},
				) ;
		} elsif ( defined $colarr_ref->{$name}{print_callback} ) {
			$self->{ess_params}{evv}->put (
				-sort_selected_event_value_callback => $colarr_ref->{$name}{print_callback},
				) ;
		}
		if ( defined $colarr_ref->{$name}{sort_type} ) {
			$self->{ess_params}{evv}->put (
				-sort_selected_event_type => $colarr_ref->{$name}{sort_type},
				) ;
		} else {
			$self->{ess_params}{evv}->put (
				-sort_selected_event_type => 'string',
				) ;
		}
	}

	if (defined $reversed) {
		$self->{ess_params}{evv}->put (
			-sort_selected_event_reversed => $reversed,
			) ;
	}

	if (defined $name || defined $reversed) {
		$self->{ess_params}{evv}->update_selected_event ;

		($self->{ess_params}{ssrow2dbrow_selected}) = $self->{ess_params}{evv}->get ( 'selected_sortedrow2dbrow' ) ;
		($self->{ess_params}{dbrow2ssrow_selected}) = $self->{ess_params}{evv}->get ( 'selected_dbrow2sortedrow' ) ;
		($self->{ess_params}{selected_origin_dbrow}) = $self->{ess_params}{evv}->get ( 'selected_origin_dbrow' ) ;

		my $dbt = $self->Subwidget ( 'event_table' ) ;
		my $selected_dbrow = $self->{ess_params}{selected_origin_dbrow} ;

		set_origin_selection ( $self, $dbt, $selected_dbrow ) ;

		$dbt->configure (
			-state => 'normal',
			) ;
		$dbt->update ;
	}
}

sub sort_event {
	my $self = shift ;
	my $value = shift ;

	if ( ! defined $value ) {
		$value = $self->{ess_params}{sort_event} ;
		return $value ;
	}

	$self->{ess_params}{sort_event} = $value ;

	if ( ! defined $self->{ess_params}{evv} ) {
		# print "In sort_event with undef evv\n" ;
		$self->{ess_params}{sort_event_need} = 1 ;
		return ;
	}
	undef $self->{ess_params}{sort_event_need} ;
	# print "In sort_event sorting on $value\n" ;

	sort_event_private ( $self ) ;

	if (defined $self->cget(-coldisplay_event)) {
		&setup_title_buttons ( $self, $self->cget(-coldisplay_event), $self->cget(-col_event), 
					\$self->{ess_params}{sort_event},
					\$self->{ess_params}{sort_event_reversed},
					$self->{ess_params}{col_event_buttons},
					\&sort_event_private ) ;
	}
}

sub sort_event_reversed {
	my $self = shift ;
	my $value = shift ;

	if ( ! defined $value ) {
		$value = $self->{ess_params}{sort_event_reversed} ;
		return $value ;
	}

	$self->{ess_params}{sort_event_reversed} = $value ;

	if ( ! defined $self->{ess_params}{evv} ) {
		# print "In sort_event_reversed with undef evv\n" ;
		$self->{ess_params}{sort_event_reversed_need} = 1 ;
		return ;
	}
	undef $self->{ess_params}{sort_event_reversed_need} ;
	# print "In sort_event_reversed sorting on $value\n" ;

	sort_event_private ( $self ) ;
}

sub setcolwidth {
	my $dbt = shift ;
	my $col = shift ;
	my $val = shift ;
	my $colwidths_ref = shift ;

	my $w = length ( $val ) ;

	my @colwidths = @{$colwidths_ref} ;

	if ( ! defined $colwidths[$col] || $w > $colwidths[$col] ) {
		$dbt->colWidth ( $col, $w ) ;
		$colwidths[$col] = $w ;
	}

	return @colwidths ;
}

sub coldisplay_events {
	my $self = shift ;
	my $value = shift ;

	if ( ! defined $value ) {
		$value = $self->{ess_params}{coldisplay_events} ;
		return $value ;
	}

	$self->{ess_params}{coldisplay_events} = $value ;

	if ( ! defined $self->{ess_params}{evv} ) {
		# print "In coldisplay_events with undef evv\n" ;
		$self->{ess_params}{coldisplay_events_need} = 1 ;
		return ;
	}
	undef $self->{ess_params}{coldisplay_events_need} ;
	# print "In coldisplay_events with defined evv\n" ;

	my $dbt = $self->Subwidget ( 'events_table' ) ;
	my $colwidths = $self->{ess_params}{colwidths_events} ;

	my $n = 0 ;
	if ( defined $self->{ess_params}{evv} ) {
		$n = @{$self->{ess_params}{evv}->{sortedrow2dbrow}} ;
	}
	# print $n, " events to display\n" ;

	&setup_title ( $value, $self->cget(-col_events), $dbt, $n, $colwidths, "#00f0f0",
				$self->{ess_params}{col_events_buttons} ) ;

	&setup_title_buttons ( $self, $value, $self->cget(-col_events), 
				\$self->{ess_params}{sort_events},
				\$self->{ess_params}{sort_events_reversed},
				$self->{ess_params}{col_events_buttons},
				\&sort_events_private ) ;
}

sub coldisplay_event {
	my $self = shift ;
	my $value = shift ;

	if ( ! defined $value ) {
		$value = $self->{ess_params}{coldisplay_event} ;
		return $value ;
	}

	$self->{ess_params}{coldisplay_event} = $value ;

	if ( ! defined $self->{ess_params}{evv} ) {
		# print "In coldisplay_event with undef evv\n" ;
		$self->{ess_params}{coldisplay_event_need} = 1 ;
		return ;
	}
	undef $self->{ess_params}{coldisplay_event_need} ;
	# print "In coldisplay_event with defined evv\n" ;

	my $dbt = $self->Subwidget ( 'event_table' ) ;
	my $colwidths = $self->{ess_params}{colwidths_event} ;

	my $n = 0 ;
	if ( defined $self->{ess_params}{evv} ) {
		$n = @{$self->{ess_params}{evv}->{selected_sortedrow2dbrow}} ;
	}
	# print $n, " origins to display\n" ;

	&setup_title ( $value, $self->cget(-col_event), $dbt, $n, $colwidths, "#00f000",
				$self->{ess_params}{col_event_buttons} ) ;

	&setup_title_buttons ( $self, $value, $self->cget(-col_event), 
				\$self->{ess_params}{sort_event},
				\$self->{ess_params}{sort_event_reversed},
				$self->{ess_params}{col_event_buttons},
				\&sort_event_private ) ;
}

sub setup_title_buttons {
	my $self = shift ;
	my $coldisplay_ref = shift ;
	my $colarr_ref = shift ;
	my $sort_ref = shift ;
	my $sort_reversed_ref = shift ;
	my $buttons = shift ;
	my $sort_callback = shift ;

	my %colarr = %{$colarr_ref} ;
	my @coldisplay = @{$coldisplay_ref} ;
	my %colbuttons = %{$buttons} ;

	my $sort ;
	my $sort_reversed ;

	if ( defined $sort_ref ) {
		$sort = ${$sort_ref} ;
	}
	if ( defined $sort_reversed_ref ) {
		$sort_reversed = ${$sort_reversed_ref} ;
	}

	foreach my $name ( @coldisplay ) {
		my $arrow = 'arrownone' ;
		if ( defined $sort && $sort eq $name && ! $sort_reversed ) {
			$arrow = 'arrowup' ;
		} elsif ( defined $sort && $sort eq $name && $sort_reversed ) {
			$arrow = 'arrowdown' ;
		}
		$colbuttons{$name}->configure (
			-bitmap => $arrow,
			) ;
		$colbuttons{$name}->bind ( '<ButtonPress-1>', [ sub {
					my $w = shift ;
					my $self = shift ;
					my $name = shift ;
					my $colbuttons_ref = shift ;
					my $coldisplay_ref = shift ;
					my $colarr_ref = shift ;
					my $sortname_ref = shift ;
					my $sortreversed_ref = shift ;
					my $sort_callback = shift ;

					if ($name eq ${$sortname_ref}) {
						if ( ${$sortreversed_ref} ) {
							${$sortname_ref} = "" ;
							${$sortreversed_ref} = 0 ;
						} else {
							${$sortreversed_ref} = 1 ;
						}
					} else {
						${$sortname_ref} = $name ;
						${$sortreversed_ref} = 0 ;
					}
					foreach my $cname ( @{$coldisplay_ref} ) {
						my $arrow = 'arrownone' ;
						if ( ${$sortname_ref} eq $cname && ! ${$sortreversed_ref} ) {
							$arrow = 'arrowup' ;
						} elsif ( ${$sortname_ref} eq $cname && ${$sortreversed_ref} ) {
							$arrow = 'arrowdown' ;
						}
						$colbuttons_ref->{$cname}->configure (
							-bitmap => $arrow,
							) ;
					}
					&{$sort_callback} ( $self ) ;
				}, $self, $name, $buttons, $coldisplay_ref, $colarr_ref, $sort_ref, $sort_reversed_ref, $sort_callback],
			) ;
	}

}

sub setup_title {
	my $value = shift ;
	my $colarr_ref = shift ;
	my $dbt = shift ;
	my $nrows = shift ;
	my $colwidths = shift ;
	my $title_color = shift ;
	my $buttons = shift ;

	my %colarr = %{$colarr_ref} ;

	my @coldisplay = @{$value} ;

	$dbt->configure (
		-rows	=> $nrows + 1,
		-cols	=> ( @coldisplay ) + 1,
		) ;

	my %colbuttons ;

	my $i=0 ;
	foreach my $name ( @coldisplay ) {
		$i++ ;
		$dbt->tagCol ( $name, $i ) ;
		$dbt->tagConfigure ( $name,
 			-anchor => $colarr{$name}{anchor},
 			) ;
	
		$colbuttons{$name} = $dbt->Button (
			-compound => 'right',
			-background => $title_color,
			-text => $colarr{$name}{label},
			-bitmap => 'arrownone',
			-anchor => $colarr{$name}{anchor},
			-relief => 'flat',
			) ;
		my $menu = $colbuttons{$name}->Menu (
			-tearoff => 0,
			-popover => $colbuttons{$name},
			-menuitems => [
					[Command => $colarr{$name}{description}
						],
					"",
					[Command => 'Delete', -command => [ sub {
							my $col = shift ;
							my $name = shift ;
							delete_col ($dbt, $name, \@coldisplay) ;
						}, $i, $name] ],
				],
			) ;
		$dbt->windowConfigure ( '0,' . $i, 
			-window => $colbuttons{$name},
			-sticky => 'nsew',
 			-background => "#fff0e0",
			) ;
		@{$colwidths} = setcolwidth ( $dbt, $i, $colarr{$name}{width_template}, $colwidths ) ;
	
		$colbuttons{$name}->bind ( '<Button-3>', [sub { 
				my ($w) = @_; 
				my $e = $w->XEvent ;
				$menu->Post ($e->X, $e->Y ) ; } ] ) ;
	}

	if (defined $buttons) {%{$buttons} = %colbuttons;}
}

sub getval {
	my @db ;
	$db[0] = shift ;
	$db[1] = shift ;
	$db[2] = shift ;
	$db[3] = shift ;
	my $params_ref = shift ;

	my %params = %{$params_ref} ;

	my $val = "" ;
	if (ref($params{print_callback}) eq "CODE") {
		$val = $params{print_callback} ( @db ) ;
	} elsif (ref($params{print_callback}) eq "ARRAY") {
		my @args = @{$params{print_callback}} ;
		my $proc = shift @args ;
		$val = &{$proc} ( @db, @args ) ;
	} else {
		$val = trim ( dbex_eval ( @db, $params{print_callback} ) ) ;
	}

	return $val ;
}

sub getcell {
	my $row = shift ;
	my $col = shift ;
	my $db_ref = shift ;
	my $coldisplay = shift ;
	my $colarr = shift ;
	my $ssrow2dbrow_ref = shift ;

	my @db = @{$db_ref} ;
	my @ssrow2dbrow = @{$ssrow2dbrow_ref} ;

	my $val ;
	if ($row == 0) {
		return "" ;
	}
	if ($col == 0) {
		$val = sprintf "%d", $row-1 ;
		return $val ;
	}
	my %cellparams = %{$colarr->{$coldisplay->[$col-1]}} ;
	if ( ! @ssrow2dbrow ) { return "" ; }
	
	@db = dblookup ( @db, 0, 'origin', 0, 0 ) ;
	$db[3] = $ssrow2dbrow[$row-1] ;
	if ( ! $db[3] ) { return "" ; }

	$val = getval ( @db, \%cellparams ) ;

	return " " . $val . " " ;
}

sub cellcmd_events {
	my $self = shift ;
	my $table = shift ;
	my $row = shift ;
	my $col = shift ;

	my $val = getcell ( $row, $col, $self->{ess_params}{db_ref}, 
				$self->{ess_params}{coldisplay_events},
				$self->cget(-col_events), 
				$self->{ess_params}{ssrow2dbrow_events} ) ;

	if ($row-1 == $self->{ess_params}{selected_event_sortedrow}) {
		$table->tagRow ( 'selectedevents', $row ) ;
	} else {
		if ( $row % 2 ) {
			$table->tagRow ( 'events1', $row ) ;
		} else {
			$table->tagRow ( 'events2', $row ) ;
		}
	}

	@{$self->{ess_params}{colwidths_events}} 
				= setcolwidth ( $table, $col, $val, 
				$self->{ess_params}{colwidths_events} ) ;

	return $val ;
}

sub cellcmd_event {
	my $self = shift ;
	my $table = shift ;
	my $row = shift ;
	my $col = shift ;

	my $val = getcell ( $row, $col, $self->{ess_params}{db_ref}, 
				$self->{ess_params}{coldisplay_event},
				$self->cget(-col_event), 
				$self->{ess_params}{ssrow2dbrow_selected} ) ;

	my $index = '@' . $col . "," . $row ;
	if ($row-1 == $self->{ess_params}{selected_origin_sortedrow}) {
		if ($self->{ess_params}{ssrow2dbrow_selected}->[$row-1] ==
					$self->{ess_params}{selected_event_dbrow}) {
			$table->tagRow ( 'pselectedorigins', $row ) ;
		} else {
			$table->tagRow ( 'selectedorigins', $row ) ;
		}
	} else {
		if ($self->{ess_params}{ssrow2dbrow_selected}->[$row-1] ==
					$self->{ess_params}{selected_event_dbrow}) {
			if ( $row % 2 ) {
				$table->tagRow ( 'porigins1', $row ) ;
			} else {
				$table->tagRow ( 'porigins2', $row ) ;
			}
		} else {
			if ( $row % 2 ) {
				$table->tagRow ( 'origins1', $row ) ;
			} else {
				$table->tagRow ( 'origins2', $row ) ;
			}
		}
	}

	@{$self->{ess_params}{colwidths_event}} 
				= setcolwidth ( $table, $col, $val, 
				$self->{ess_params}{colwidths_event} ) ;

	return $val ;
}

sub prettyprint {
	my $val = shift;
	my $prefix = "";
	if (@_) { $prefix = shift ; }

	if (ref($val) eq "HASH") {
		my @keys = sort ( keys  %$val );
		my %hash = %$val;
		foreach my $key (@keys) {
			my $newprefix = $prefix . "{". $key . "}" ;
			prettyprint ($hash{$key}, $newprefix) ;
		}
	} elsif (ref($val) eq "ARRAY") {
		my $i = 0;
		my @arr = @$val;
		foreach my $entry ( @$val ) {
			my $newprefix = $prefix . "[". $i . "]" ;
			prettyprint ($arr[$i], $newprefix) ;
			$i++;
		}
	} else {
		if (substr($prefix, -4, 4) eq "blob") {
			print $prefix, " = ";
			my @vals = unpack "C9", $val ;
			my $i = 0;
			foreach $val (@vals) {
				printf "%2.2X ", $val ;
				$i++ ;
				if ($i == 8) { last; }
			}
			if ( scalar (@vals) > 8 ) { print "..." ; }
			print "\n";
		} else {
			print $prefix, " = ", $val, "\n";
		}
	}
}

#package main ;
#
#use Tk ;
#use Datascope ;
#
#my $dbname = shift ;
#my @db = dbopen ( $dbname, 'r' ) ;
#
#my $mw = MainWindow->new (
#	) ;
#$mw->gridRowconfigure ( 0, -weight => 1 ) ;
#$mw->gridColumnconfigure ( 0, -weight => 1 ) ;
#
#my $event_frame = $mw->EventSpreadsheet (
#		-db => \@db,
#	)->grid (
#		-row => 0,
#		-column => 0,
#		-sticky => 'nswe',
#		) ;
#
#MainLoop ;
