/* Copyright (c) 2008 Boulder Real Time Technologies, Inc. */
/* All rights reserved */
 
/* This software module is wholly owned by Boulder Real Time 
   Technologies, Inc. This software may be used freely in any 
   way as long as the copyright statement above is not removed. */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <strings.h>

#include "import.h"
#include "Pkt.h"

static int
import_process_dbin (void *imv)

{
	Import *im = (Import *) imv;
	long isview, n;
	Tbl *keys;
	Dbptr dbg;
	double tmin, tmax;
	char expr[512];

	if (im == NULL) return (0);

	dbquery (im->dbin, dbTABLE_IS_VIEW, &isview);

	if (isview) {
		Tbl *tables;
		char *table;

		dbquery (im->dbin, dbVIEW_TABLES, &tables);
		if (maxtbl(tables) != 1) {
			register_error (0, "import_process_dbin: Input view should consist of single wfdisc table only.\n");
			return (-1);
		}
		table = (char *) gettbl (tables, 0);
		if (strcmp(table, "wfdisc")) {
			register_error (0, "import_process_dbin: Input view should consist of single wfdisc table only.\n");
			return (-1);
		}
	} else {
		im->dbin = dblookup (im->dbin, 0, "wfdisc", 0, 0);
	} 

	dbquery (im->dbin, dbRECORD_COUNT, &n);
	if (n == 0) {
		register_error (0, "import_process_dbin: No rows to process in input database.\n");
		return (-1);
	}

	for (im->dbin.record=0; im->dbin.record<n; im->dbin.record++) {
		double time, endtime;

		dbgetv (im->dbin, 0, "time", &time, "endtime", &endtime, NULL);
		if (im->dbin.record == 0) {
			tmin = time;
			tmax = endtime;
			continue;
		} 
		if (time < tmin) tmin = time;
		if (endtime > tmax) tmax = endtime;
	}

	if (im->starttime == 0.0) im->starttime = tmin-im->backup;
	if (im->endtime == 0.0) im->endtime = tmax;

	sprintf (expr, "time <= %lf && endtime >= %lf", im->endtime, im->starttime);
	im->dbin = dbsubset (im->dbin, expr, 0);

	dbquery (im->dbin, dbRECORD_COUNT, &n);
	if (n == 0) {
		register_error (0, "import_process_dbin: No rows to process in input database.\n");
		return (-1);
	}

	keys = strtbl ( "sta", "chan", "time", NULL );
	im->dbin = dbsort (im->dbin, keys, 0, 0);
	freetbl (keys, NULL);

	keys = strtbl ( "sta", "chan", NULL );
	dbg = dbgroup (im->dbin, keys, 0, 1);
	freetbl (keys, NULL);

	dbquery (dbg, dbRECORD_COUNT, &n);
	im->channels = (struct channels_ *) malloc (sizeof(struct channels_)*n);
	if (im->channels == NULL) {
		register_error (1, "import_process_dbin: malloc(channels,%ld) error.\n", sizeof(struct channels_)*n);
		return (-1);
	}
	im->nchannels = n;
	memset (im->channels, 0, sizeof(struct channels_)*n);

	for (dbg.record=0; dbg.record<im->nchannels; dbg.record++) {
		Dbptr bundle;
		long is, ie;

		if (dbgetv (dbg, 0, 	"sta", im->channels[dbg.record].sta, 
					"chan", im->channels[dbg.record].chan, 
					"bundle", &bundle, 
					NULL) < 0) {
			register_error (0, "import_process_dbin: dbgetv() error.\n");
			return (-1);
		}
		dbget_range (bundle, &is, &ie);
		im->channels[dbg.record].rec = is;
		im->channels[dbg.record].erec = ie;
		im->dbin.record = is;
		dbseed_net (im->dbin, im->channels[dbg.record].sta, im->channels[dbg.record].snet, im->channels[dbg.record].ssta);
		clear_register (0);
		dbseed_loc (im->dbin, im->channels[dbg.record].sta, im->channels[dbg.record].chan, im->channels[dbg.record].schan, im->channels[dbg.record].sloc);
		clear_register (0);
	}

	im->dbin.record = 0;

	return (0);
}

static int 
import_getcalib (Import *im, char *sta, char *chan, double time, double endtime, double *calib, double *calper, char *segtype)

{
	Tbl *tbl=NULL;
	Dbptr dbc, dbm;
	int ret;

	if (im->calibfromdb == 0) return (0);

	dbc = dblookup (im->dbin, 0, "calibration", 0, "dbSCRATCH");
	dbm = dblookup (im->dbin, 0, "calibration", 0, 0);
	dbputv (dbc, 0, "sta", sta, "chan", chan, "time", time, "endtime", endtime, NULL );
	ret = dbmatches (dbc, dbm, NULL, NULL, &(im->calib_hook), &tbl);
	if (ret < 0) {
		wfproc_mutex_unlock ("db");
		register_error (0, 
			"import_getcalib: dbmatches(%s,%s,calibration) error.\n", sta, chan);
		return (-1);
	}
	if (ret >= 1) {
		dbm.record = (long) gettbl (tbl, 0);
		if (dbgetv (dbm, 0,	"calib", calib,
					"calper", calper,
					"segtype", segtype,
					NULL ) < 0) {
			freetbl (tbl, NULL);
			wfproc_mutex_unlock ("db");
			register_error (0, 
				"import_getcalib: dbgetv(%s,%s,calib,...) error.\n", sta, chan);
			return (-1);
		}
	}
	freetbl (tbl, NULL);

	return (0);
}

static int 
import_getcalib_for_pktchan (Import *im, PktChannel *pktchan)

{
	char sta[32], chan[32];
	int ret;

	if (im->calibfromdb == 0) return (0);

	if (wfprocdb_lock (im->wfpdbin, &(im->dbin)) < 0) {
		register_error (0, "import_getcalib_for_pktchan: wfprocdb_lock() error.\n");
		return (-1);
	}

	dbmap_seed_netsta (im->dbin, pktchan->net, pktchan->sta, sta);
	clear_register (1);
	dbmap_seed_chanloc (im->dbin, sta, pktchan->chan, pktchan->loc, chan);
	clear_register (1);

	ret = import_getcalib (im, sta, chan, pktchan->time, pktchan->time+(pktchan->nsamp-1)/pktchan->samprate, 
							&(pktchan->calib), &(pktchan->calper), pktchan->segtype);

	wfprocdb_unlock(im->wfpdbin);


	return (ret);
}

void *
import_new (Pf *params)

{
	size_t size;
	Import *im;

	size = sizeof(Import);
	im = (Import *) malloc (size);
	if (im == NULL) {
		register_error (1, "import_new: malloc(Import,%ld) error.\n", size);
		return (NULL);
	}

	memset (im, 0, size);

	im->dbin = dbinvalid();
	im->orb = -1;
	im->told = now();
	trfill_gap (&(im->gap_value), 1);

	return (im);
}

int
import_destroy (void *imv)

{
	Import *im = (Import *) imv;

	if (imv == NULL) return (0);

	if (im->orb >= 0) orbclose (im->orb);

	if (im->channels) free (im->channels);

	if (im->pt) pktchannel2trace_destroy (im->pt);

	free (imv);

	return (0);
}

int
import_config (void *imv, Pf *params, Pf *state)

{
	Import *im = (Import *) imv;
	char input[512];
	int pktid;
	char *ptr;
	WFProcDb *dbi;
	char strbuf[512];
	int ret;

	strcpy (im->start, "");
	if (parse_param (params, "-start", P_STRCP, 0, im->start) < 0) {
		register_error (0, "import_config: parse_param(-start) error.\n");
		return (-1);
	}
	if (!strcmp(im->start, "")) {
		if (parse_param (params, "start", P_STRCP, 0, im->start) < 0) {
			register_error (0, "import_config: parse_param(start) error.\n");
			return (-1);
		}
	}
	im->backup = 0.0;
	if (parse_param (params, "backup", P_DBL, 0, &(im->backup)) < 0) {
		register_error (0, "import_config: parse_param(backup) error.\n");
		return (-1);
	}
	im->starttime = 0.0;
	if (strcmp(im->start, "")) {
		double time;

		if (!strncmp(im->start, "now-", 4)) {
			double shift;

			shift = atof(&(im->start[4]));
			time = now()-shift;
		} else {
			if (zstr2epoch (im->start, &time) < 0) {
				register_error (0, "import_config: Unable to parse start time string '%s'.\n",
										im->start);
				return (-1);
			}
			time -= im->backup;
		}
		im->starttime = time;
	}

	strcpy (im->end, "");
	if (parse_param (params, "-end", P_STRCP, 0, im->end) < 0) {
		register_error (0, "import_config: parse_param(-end) error.\n");
		return (-1);
	}
	if (!strcmp(im->end, "")) {
		if (parse_param (params, "end", P_STRCP, 0, im->end) < 0) {
			register_error (0, "import_config: parse_param(end) error.\n");
			return (-1);
		}
	}
	if (strcmp(im->end, "")) {
		double time;

		if (zstr2epoch (im->end, &time) < 0) {
			register_error (0, "import_config: Unable to parse end time string '%s'.\n",
									im->end);
			return (-1);
		}
		im->endtime = time;
	}

	strcpy (input, "");
	if (parse_param (params, "heap_throttle", P_STRCP, 0, input) < 0) {
		register_error (0, "import_config: parse_param(heap_throttle) error.\n");
		return (-1);
	}
	if (*input) {
		size_t mult = 1;

		if (input[strlen(input)-1] == 'G') {
			mult = 1000000000;
			input[strlen(input)-1] = '\0';
		}
		if (input[strlen(input)-1] == 'M') {
			mult = 1000000;
			input[strlen(input)-1] = '\0';
		}
		if (input[strlen(input)-1] == 'K') {
			mult = 1000;
			input[strlen(input)-1] = '\0';
		}
		im->heap_throttle = atoi(input)*mult;
	}
	if (parse_param (params, "-verbose", P_LINT, 0, &(im->verbose)) < 0) {
		register_error (0, "import_config: parse_param(-verbose) error.\n");
		return (-1);
	}
	if (parse_param (params, "verbose", P_LINT, 0, &(im->verbose)) < 0) {
		register_error (0, "import_config: parse_param(verbose) error.\n");
		return (-1);
	}

	im->calibfromdb = 0;
	if (parse_param (params, "calibfromdb", P_BOOL, 0, &(im->calibfromdb)) < 0) {
		register_error (0, "import_config: parse_param(calibfromdb) error.\n");
		return (-1);
	}

	strcpy (im->input_type, "orb");
	if (parse_param (params, "input_type", P_STRCP, 0, im->input_type) < 0) {
		register_error (0, "import_config: parse_param(input_type) error.\n");
		return (-1);
	}

	if (!strcmp(im->input_type, "db")) {
		char *viewname;

		if (parse_param (params, "input", P_STRCP, 1, input) < 0) {
			register_error (0, "import_config: parse_param(input) error.\n");
			return (-1);
		}
	
		if (parse_param (params, input, P_STRCP, 1, im->dbname) < 0) {
			register_error (0, "import_config: parse_param(%s) error.\n", input);
			return (-1);
		}

		viewname = NULL;
		if (parse_param (params, "-i", P_STR, 0, &viewname) < 0) {
			register_error (0, "import_config: parse_param(-i) error.\n");
			return (-1);
		}
		im->pause_timeout = 1800.0;
		if (parse_param (params, "-pause_timeout", P_DBL, 0, &(im->pause_timeout)) < 0) {
			register_error (0, "import_config: parse_param(-pause_timeout) error.\n");
			return (-1);
		}

		im->wfpdbin = wfprocdb_create (im->dbname, viewname, 1, im->pause_timeout);
		if (im->wfpdbin == NULL) {
			if (viewname) {
				register_error (0, "import_config: wfprocdb_create (%s,%s) error.\n", im->dbname, viewname);
			} else {
				register_error (0, "import_config: wfprocdb_create (%s) error.\n", im->dbname);
			}
			return (-1);

		}
		if (wfprocdb_lock (im->wfpdbin, &(im->dbin)) < 0) {
			register_error (0, "import_config: wfprocdb_lock() error.\n");
			return (-1);
		}
		wfprocdb_unlock (im->wfpdbin);

		if (import_process_dbin(imv) < 0) {
			register_error (0, "import_config: import_process_dbin() error.\n");
			return (-1);
		}
		if (parse_param (params, "time_slice_duration", P_DBL, 1, &(im->time_slice_duration)) < 0) {
			register_error (0, "import_config: parse_param(time_slice_duration) error.\n");
			return (-1);
		}

		return (0);
	} else if (strcmp(im->input_type, "orb")) {
		register_error (0, "import_config: Unrecignized input_type, %s.\n", im->input_type);
		return (-1);
	}

	if (im->orb >= 0) orbclose (im->orb);

	ptr = NULL;
	if (parse_param (params, "dbmaster", P_STR, 0, &ptr) < 0) {
		register_error (0, "import_config: parse_param(dbmaster) error.\n");
		return (-1);
	}
	if (ptr) {
		ret = sscanf (ptr, "%lu", (unsigned long *) &dbi);
		free (ptr);
		if (ret != 1) {
			register_error (0, "import_config: Cannot parse 'dbmaster' parameter (%s).\n", strbuf);
			return (-1);
		}
		im->wfpdbin = dbi;
	}

	if (parse_param (params, "chan_accept", P_TBL, 0, &(im->chan_accept)) < 0) {
		register_error (0, "import_config: parse_param(chan_accept) error.\n");
		return (-1);
	}
	if (parse_param (params, "chan_reject", P_TBL, 0, &(im->chan_reject)) < 0) {
		register_error (0, "import_config: parse_param(chan_reject) error.\n");
		return (-1);
	}

	if (parse_param (params, "too_old", P_DBL, 0, &(im->too_old)) < 0) {
		register_error (0, "import_config: parse_param(too_old) error.\n");
		return (-1);
	}
	if (parse_param (params, "too_new", P_DBL, 0, &(im->too_new)) < 0) {
		register_error (0, "import_config: parse_param(too_new) error.\n");
		return (-1);
	}

	if (parse_param (params, "input", P_STRCP, 1, input) < 0) {
		register_error (0, "import_config: parse_param(input) error.\n");
		return (-1);
	}

	if (parse_param (params, input, P_STRCP, 1, im->orbname) < 0) {
		register_error (0, "import_config: parse_param(%s) error.\n", input);
		return (-1);
	}

	im->orb = orbopen (im->orbname, "r&");
	if (im->orb < 0) {
		register_error (0, "import_config: orbopen(%s) error.\n", im->orbname);
		return (-1);
	}

	strcpy (im->select, "");
	if (parse_param (params, "-select", P_STRCP, 0, im->select) < 0) {
		register_error (0, "import_config: parse_param(-select) error.\n");
		return (-1);
	}
	if (!strcmp(im->select, "")) {
		if (parse_param (params, "select", P_STRCP, 0, im->select) < 0) {
			register_error (0, "import_config: parse_param(select) error.\n");
			return (-1);
		}
	}
	if (strcmp(im->select, "")) {
		if (orbselect (im->orb, im->select) < 0) {
			register_error (0, "import_config: orbselect(%s,%s) error.\n", im->orbname, im->select);
			return (-1);
		}
	}

	strcpy (im->reject, "");
	if (parse_param (params, "-reject", P_STRCP, 0, im->reject) < 0) {
		register_error (0, "import_config: parse_param(-reject) error.\n");
		return (-1);
	}
	if (!strcmp(im->reject, "")) {
		if (parse_param (params, "reject", P_STRCP, 0, im->reject) < 0) {
			register_error (0, "import_config: parse_param(reject) error.\n");
			return (-1);
		}
	}
	if (strcmp(im->reject, "")) {
		if (orbreject (im->orb, im->reject) < 0) {
			register_error (0, "import_config: orbreject(%s,%s) error.\n", im->orbname, im->reject);
			return (-1);
		}
	}
	im->process_through_unstuff_errors = 0;
	if (parse_param (params, "process_through_unstuff_errors", P_BOOL, 0, &(im->process_through_unstuff_errors)) < 0) {
		register_error (0, "import_config: parse_param(process_through_unstuff_errors) error.\n");
		return (-1);
	}

	pktid = -1;
	if (state) {
		int lastpktid;
		double lastpkttime;

		if (im->backup > 0.0) {
			if (parse_param (state, "lastpkttime", P_DBL, 1, &lastpkttime) < 0) {
				complain (0, "import_config: parse_param(lastpkttime) error....using start parameters\n");
			} else {
				im->starttime = lastpkttime-im->backup;
			}
		} else {
			if (parse_param (state, "lastpktid", P_LINT, 1, &lastpktid) < 0) {
				complain (0, "import_config: parse_param(lastpktid) error....using start parameters\n");
			} else {
				pktid = orbseek (im->orb, lastpktid);
				if (pktid < 0) {
					complain (0, "import_config: orbseek(lastpktid) error...using start parameters.\n");
					pktid = -1;
				} else {
					complain (0, "import_config: resurrection successfull...start following pktid %d.\n", pktid);
				}
			}
		}
	}

	if (pktid == -1 && im->starttime > 0.0) {
		char *s;

		s = strtime(im->starttime);
		elog_notify (0, "import_config: calling orbafter(%s,%s)\n", im->orbname, s);
		free (s);
		if (orbafter (im->orb, im->starttime) < 0) {
			s = strtime(im->starttime);
			complain (0, "import_config: orbafter(%s,%s) error...starting at next packet\n", im->orbname, s);
			free (s);
		}
	}

	im->orbth = orbreapthr_new (im->orb, 0.1, 1);

	return (0);
}

int
import_proc_make_time_slice (void *imv, Packet **pkt)

{
	Import *im = (Import *) imv;
	int ret=0, i, j, nchannels;
	long n;

	if (*pkt) freePkt (*pkt);
	*pkt = NULL;

	if (im->time_slice_duration > 0.0) {
		if (im->time_slice_time == 0.0) {
			im->time_slice_time = im->starttime;
		} else {
			im->time_slice_time = im->time_slice_endtime;
		}
		im->time_slice_endtime = im->time_slice_time + im->time_slice_duration;
		if (im->time_slice_time > im->endtime) return (1);
		if (im->time_slice_endtime > im->endtime) im->time_slice_endtime = im->endtime;
		if (im->time_slice_endtime <= im->time_slice_time) return (1);

		if (im->verbose > 2) {
			char *s;
	
			s = strtime(im->time_slice_time);
			elog_notify (0, "import_proc_make_time_slice: Processing time slice %s\n", s);
			free (s);
		}
		nchannels = im->nchannels;
	} else {
		double samprate;
		long nsamp;
		double ts, te;
		PktChannel *pchan;
		float *fdata;
		float *data = NULL;
		long datasz = 0;
		char sta[32], chan[32];
		char snet[32], ssta[32], schan[32], sloc[32];
		double calib, calper;
		char segtype[32];

		wfproc_mutex_lock ("db");

		dbquery (im->dbin, dbRECORD_COUNT, &n);
		if (im->dbin.record >= n) {
			wfproc_mutex_unlock ("db");
			return (1);
		}
		if (dbgetv (im->dbin, 0,	"sta", sta,
						"chan", chan,
						"time", &(im->time_slice_time),
						"endtime", &(im->time_slice_endtime),
						"samprate", &samprate,
						"calib", &calib,
						"calper", &calper,
						"segtype", segtype,
						NULL) < 0) {
			register_error (0, "import_proc_make_time_slice: dbgetv() error.\n");
			wfproc_mutex_unlock ("db");
			return (-1);
		}

		im->time_slice_endtime += 1.0/samprate;
		if (im->time_slice_time > im->endtime) {
			wfproc_mutex_unlock ("db");
			return (0);
		}
		if (im->time_slice_endtime > im->endtime) im->time_slice_endtime = im->endtime;
		if (im->time_slice_endtime <= im->time_slice_time) {
			wfproc_mutex_unlock ("db");
			return (0);
		}

		if (import_getcalib (im, sta, chan, im->time_slice_time, im->time_slice_endtime, &calib, &calper, segtype) < 0) {
			register_error (0, "import_proc_make_time_slice: import_getcalib(%s,%s) error.\n", sta, chan);
			wfproc_mutex_unlock ("db");
			return (-1);
		}

		ret = trgetwf (im->dbin, 0, &data, &datasz, im->time_slice_time, im->time_slice_endtime-0.5/samprate, &ts, &te, &nsamp, 0, 0);
		im->dbin.record++;

		wfproc_mutex_unlock ("db");

		if (ret < 0) {
			trgetwf_error (im->dbin, ret);
			register_error (0, "import_proc_time_slice: trgetwf(%s,%s) error.\n", sta, chan);
			return (-1);
		}
		if (nsamp < 1) return (0);

		pchan = newPktChannel ();
		if (pchan == NULL) {
			register_error (0, "import_proc_time_slice: newPktChannel(pchan) error.\n");
			return (-1);
		}
		pchan->data = (int *)((float *) malloc (nsamp*sizeof(float)));
		if (pchan->data == NULL) {
			register_error (0, "pktchannel2trace_put_from_db: malloc(pchan->data,%ld) error.\n",
										nsamp*sizeof(int));
			return (-1);
		}
		pchan->datasz = (int)nsamp;
		memset (pchan->data, 0, pchan->datasz*sizeof(float));
		fdata = (float *) pchan->data;

		for (j=0; j<nsamp; j++) {
			fdata[j] = data[j] ;
		}
		free (data);
		pchan->nsamp = (int)nsamp;
		pchan->time = ts;
		pchan->samprate = samprate;
		pchan->calib = calib;
		pchan->calper = calper;
		pchan->isfloat = 1;
		strcpy (pchan->segtype, segtype);

		dbseed_net (im->dbin, sta, snet, ssta);
		dbseed_loc (im->dbin, sta, chan, schan, sloc);
		clear_register (0);
		strcpy (pchan->net, snet);
		strcpy (pchan->sta, ssta);
		strcpy (pchan->chan, schan);
		strcpy (pchan->loc, sloc);

		*pkt = newPkt ();
		if (*pkt == NULL) {
			register_error (0, "import_proc_time_slice: newPkt(pkt) error.\n");
			freePktChannel (pchan);
			return (-1);
		}

		settbl ((*pkt)->channels, -1, pchan);
		(*pkt)->nchannels = 1;

		if (im->verbose > 2) {
			char *s;
	
			s = strtime(im->time_slice_time);
			elog_notify (0, "import_proc_make_time_slice: Processing %s:%s at time %s\n", sta, chan, s);
			free (s);
		}

		return (0);
	}

	wfproc_mutex_lock ("db");
	im->dbtr = trnew(0, 0);
	im->pt = pktchannel2trace_new (im->dbin, im->dbtr, im->dbin, im->calibfromdb, 0, 0);
	if (im->pt == NULL) {
		register_error (0, "import_proc_make_time_slice: pktchannel2trace_new() error.\n");
		wfproc_mutex_unlock ("db");
		return (-1);
	}

	for (i=0; i<nchannels; i++) {
		int nrec, irec;

		nrec = pktchannel2trace_put_from_db (im->pt, im->channels[i].sta, 
				im->channels[i].chan, 1, im->time_slice_time, 
				im->time_slice_endtime);
		if (nrec < 0) {	
			register_error (0, "import_proc_make_time_slice: pktchannel2trace_put_from_db(%s,%s) error.\n", 
								im->channels[i].sta, im->channels[i].chan);
			ret = -1;
			goto OUT;
		}

		if (nrec == 0) continue;

		if (*pkt == NULL) {
			*pkt = newPkt ();
			if (*pkt == NULL) {
				register_error (0, "import_proc_make_time_slice: newPkt() error.\n");
				ret = -1;
				goto OUT;
			}
		}

		irec = im->pt->dbt.record;

		for (im->pt->dbt.record=irec; im->pt->dbt.record<irec+nrec; (im->pt->dbt.record)++) {
			PktChannel *chan;
			char stac[32], chanc[32], segtype[16];
			double time, endtime, samprate, calib, calper;
			long nsamp;
			float *data;
			int ok;

			if (dbgetv (im->pt->dbt,0,	"sta", stac, 
							"chan", chanc, 
							"time", &time,
							"nsamp", &nsamp,
							"samprate", &samprate,
							"calib", &calib, 
							"calper", &calper, 
							"segtype", segtype, 
							"data", &data, 
							NULL) < 0) {
				register_error (0, "import_proc_make_time_slice: dbgetv() error.\n");
				ret = -1;
				goto OUT;
			}
			ok = 0;
			for (j=0; j<nsamp; j++) {
				if (data[j] != im->gap_value) {
					ok = 1;
					break;
				}
			}
			if (!ok) {
				free (data);
				dbputv (im->pt->dbt, 0, "data", NULL, NULL);
				continue;
			}
			endtime = time + (nsamp-1)/samprate;
			if (endtime >= im->time_slice_time+im->time_slice_duration) {
				nsamp--;
				endtime = time + (nsamp-1)/samprate;
			}
			chan = newPktChannel();
			if (chan == NULL) {
				register_error (0, "import_proc_make_time_slice: newPktChannel() error.\n");
				ret = -1;
				goto OUT;
			}
			dbseed_net (im->dbin, stac, chan->net, chan->sta);
			clear_register (0);
			dbseed_loc (im->dbin, stac, chanc, chan->chan, chan->loc);
			clear_register (0);
			chan->time = time;
			chan->nsamp = nsamp;
			chan->samprate = samprate;
			chan->calib = calib;
			chan->calper = calper;
			strcpy (chan->segtype, segtype);
			chan->isfloat = 1;
			chan->data = (int *) data;
	
			dbputv (im->pt->dbt, 0, "data", NULL, NULL);
	
			settbl ((*pkt)->channels, -1, chan);
		}
	}	
	if (*pkt) (*pkt)->nchannels = maxtbl((*pkt)->channels);

OUT:	dbquery (im->pt->dbt, dbRECORD_COUNT, &n);

	for (im->pt->dbt.record=0; im->pt->dbt.record<n; im->pt->dbt.record++) {
		float *data;

		dbgetv (im->pt->dbt, 0, "data", &data, NULL);
		if (data) free (data);
		dbputv (im->pt->dbt, 0, "data", NULL, NULL);
	}

	pktchannel2trace_destroy (im->pt);
	wfproc_mutex_unlock ("db");
	im->pt = NULL;

	return (ret);
}

int
import_proc (void *imv, WFProcData *input, Tbl **output, Pf **state)

{
	Import *im = (Import *) imv;
	WFProcTask *this_task;
	int pktid, nbytes;
	int bufsize=0;
	char *pktbuf=NULL;
	double pkttime;
	char srcname[ORBSRCNAME_SIZE];
	int i, ret;
	Packet *pkt=NULL;
	double t0 = 0.0;

	if (input != NULL) {
		register_error (0, "import_proc: Wrong input type.\n");
		return (-1);
	}

	if (wfproc_stop) return (WFPROC_FINISHED);

	this_task = wfproc_task_get_from_obj (imv);

	if (im->orb >= 0) {
		while (1) {
			if (wfproc_stop) return (WFPROC_FINISHED);
			ret = orbreapthr_get (im->orbth, &pktid, srcname, &pkttime,
						&pktbuf, &nbytes, &bufsize);
			if (wfproc_stop) return (WFPROC_FINISHED);
			if (ret != ORBREAPTHR_OK) continue;
			if (ret < 0) {
				if (pktbuf) free (pktbuf);
				register_error (0, "import_proc: orbreap() error.\n");
				return (-1);
			}
	
			if (im->verbose > 2) {
				char *s;
	
				elog_notify (0, "%s: import_proc: read packet %s %d at %s\n", this_task->name, srcname, pktid, s=strtime(pkttime));
				free (s);
			}
	
			ret = unstuffPkt (srcname, pkttime, pktbuf, nbytes, &pkt);
			if (ret == Pkt_wf) break;
			if (ret == Pkt_float) break;
			if (ret < 0) {
				char *s;
	
				if (im->process_through_unstuff_errors) {
					complain (0, "%s: import_proc: packet %s %d at %s - unstuffPkt() error - discarding packet\n", this_task->name, srcname, pktid, s=strtime(pkttime));
					free (s);

					if (pktbuf) free (pktbuf);
					*output = newtbl (0);
					if (*output == NULL) {
						register_error (0, "import_proc: newtbl(output) error.\n");
						return (-1);
					}

					return (0);
				} else {
					if (pktbuf) free (pktbuf);
					register_error (0, "%s: import_proc: packet %s %d at %s - unstuffPkt() error - error exit\n", this_task->name, srcname, pktid, s=strtime(pkttime));
					free (s);
					return (-1);
				}
			}
		}
		if (pktbuf) free (pktbuf);

		if (im->too_old > 0.0 && now()-pkttime > im->too_old) {
			char *s;
	
			complain (0, "%s: import_proc: packet %s %d at %s too old - discarding packet\n", this_task->name, srcname, pktid, s=strtime(pkttime));
			free (s);

			if (pkt) freePkt (pkt);
			*output = newtbl (0);
			if (*output == NULL) {
				register_error (0, "import_proc: newtbl(output) error.\n");
				return (-1);
			}

			return (0);
		}

		if (im->too_new > 0.0 && pkttime-now() > im->too_new) {
			char *s;
	
			complain (0, "%s: import_proc: packet %s %d at %s too new - discarding packet\n", this_task->name, srcname, pktid, s=strtime(pkttime));
			free (s);

			if (pkt) freePkt (pkt);
			*output = newtbl (0);
			if (*output == NULL) {
				register_error (0, "import_proc: newtbl(output) error.\n");
				return (-1);
			}

			return (0);
		}

		if (im->endtime > 0.0 && pkttime > im->endtime) {
			if (pkt) freePkt (pkt);
			return (WFPROC_FINISHED);
		}

		if (pkt && im->calibfromdb) {
			for (i=0; i<pkt->nchannels; i++) {
				PktChannel *chan;

				chan = (PktChannel *) gettbl (pkt->channels, i);
				if (import_getcalib_for_pktchan (im, chan) < 0) {
					clear_register (0);
				}
			}
		}
	} else {
		if (im->endtime > 0.0 && im->time_slice_time > im->endtime) {
			return (WFPROC_FINISHED);
		}
		ret = import_proc_make_time_slice (imv, &pkt);
		if (ret < 0) {
			register_error (0, "import_proc: import_proc_make_time_slice() error.\n");
			if (pkt) freePkt (pkt);
			return (-1);
		}
		if (ret == 1) {
			return (WFPROC_FINISHED);
		}
	}

	if (wfproc_stop) return (WFPROC_FINISHED);

	if (im->heap_throttle > 0) {
		if (now()-t0 > 5.0) {
			i = 0;

			while (1) {
				size_t size;

				size = wfproc_get_fifo_size ();
				if (size < im->heap_throttle) break;
				if (im->verbose > 2) {
					if (!(i%10)) elog_notify (0, "import_proc: waiting for fifo heap total of %ld to get less than %ld.\n", size, im->heap_throttle);
					i++;
				}
				sleep (1);
			}
			t0 = now();
		}
	}

	*output = newtbl (0);
	if (*output == NULL) {
		if (pkt) freePkt (pkt);
		register_error (0, "import_proc: newtbl(output) error.\n");
		return (-1);
	}

	if (pkt == NULL) return (0);

	for (i=0; i<pkt->nchannels; i++) {
		PktChannel *chan;
		WFProcData wfpd_tmp;
		size_t size;
		WFProcData *wfpd;
		double t;
		char nscl[128];

		chan = (PktChannel *) gettbl (pkt->channels, i);
		wfproc_make_nscl (chan->net, chan->sta, chan->chan, chan->loc, nscl);
		if (im->chan_accept) {
			Tbl *accept;

			accept = tblgrep (im->chan_accept, nscl);
			if (accept == NULL || maxtbl(accept) < 1) {
				if (im->verbose) {
					elog_notify (0, "import_proc: %s: Skipping because %s doesn't match chan_accept\n", 
							this_task->name, nscl);
				}
				if (accept) freetbl (accept, NULL);
				continue;
			}
			if (accept) freetbl (accept, NULL);
		}
		if (im->chan_reject) {
			Tbl *reject;
	
			reject = tblgrep (im->chan_reject, nscl);
			if (reject != NULL && maxtbl(reject) > 0) {
				if (im->verbose) {
					elog_notify (0, "import_proc: %s: Skipping because %s does match chan_reject\n", 
							this_task->name, nscl);
				}
				freetbl (reject, NULL);
				continue;
			}
			if (reject) freetbl (reject, NULL);
		}
		size = sizeof(WFProcData);
		memset (&wfpd_tmp, 0, size);
		strcpy (wfpd_tmp.type, "wf");
		wfpd_tmp.data = chan;
		wfpd_tmp.printproc = import_data_print;
		wfpd_tmp.freeproc = import_data_free;
		wfpd_tmp.copyproc = import_data_copy;
		wfpd_tmp.topktproc = import_data_topkt;
		wfpd_tmp.size = size + sizeof(PktChannel) + 4*(chan->datasz);
		t = now();
		if (t-im->told > 300.0) {
			im->told = t;
			/* continue; */
		}
		wfpd = import_data_copy (&wfpd_tmp);
		if (wfpd == NULL) {
			register_error (0, "import_proc: import_data_copy(output) error.\n");
			return (-1);
		}
		settbl (*output, -1, wfpd);
	}
	if (pkt) freePkt (pkt);

	if (im->orb >= 0) {
		if (*state == NULL) {
			*state = pfnew (PFARR);
		}
		pfput_int (*state, "lastpktid", pktid);
		pfput_int (*state, "lastpkttime", pkttime);
	}

	return (0);
}

void
import_data_free (WFProcData *packet)

{
	PktChannel *chan;

	if (packet == NULL) return;
	chan = (PktChannel *) packet->data;
	if (chan != NULL) {
		freePktChannel (chan);
	}
	free (packet);
}

char *
import_data_print (WFProcData *packet, int verbose)

{
	PktChannel *chan;
	char line[512];
	char *s;

	if (packet == NULL) return (NULL);
	chan = (PktChannel *) packet->data;
	if (chan == NULL) return (NULL);
	s = strtime(chan->time);
	if (!strcmp(chan->loc, "") || !strcmp(chan->loc, "-")) {
		sprintf (line, "%s_%s_%s %s %d %.3f", chan->net, chan->sta, chan->chan,
				s, chan->nsamp, chan->samprate);
	} else {
		sprintf (line, "%s_%s_%s_%s %s %d %.3f", chan->net, chan->sta, chan->chan,
				chan->loc, s, chan->nsamp, chan->samprate);
	}
	free (s);

	return (strdup(line));
}

WFProcData *
import_data_copy (WFProcData *packet)

{
	WFProcData *copy;
	size_t size;
	PktChannel *chan, *chancopy;

	if (packet == NULL) return (NULL);

	size = sizeof(WFProcData);
	copy = (WFProcData *) malloc (size);
	if (copy == NULL) {
		register_error (1, "import_data_copy: malloc(WFProcData,%ld) error.\n", size);
		return (NULL);
	}

	chancopy = newPktChannel();
	if (chancopy == NULL) {
		free (copy);
		register_error (0, "import_data_copy: newPktChannel() error.\n");
		return (NULL);
	}

	chan = (PktChannel *)packet->data;
	*chancopy = *chan;
	chancopy->datasz = chancopy->nsamp;
	size = chancopy->nsamp*sizeof(int);
	chancopy->data = (int *) malloc (size);
	if (chancopy->data == NULL) {
		free (copy);
		free (chan);
		register_error (1, "import_data_copy: malloc(chancopy->data,%ld) error.\n", size);
		return (NULL);
	}

	memcpy (chancopy->data, chan->data, size);

	*copy = *packet;
	copy->data = chancopy;

	return (copy);
}

int
import_data_topkt (WFProcData *input, Tbl *translations, Tbl *pkts_tbl)

{
	PktChannel *data, *achan;
	char tnet[32], tsta[32], tchan[32], tloc[32];
	int ret;
	Packet *pkt;

	data = (PktChannel *) input->data;
	if (translations) {
		ret = wfproc_translate_seedcodes (data->net, data->sta, data->chan, data->loc,
						translations, tnet, tsta, tchan, tloc);
		if (ret < 0) {
			register_error (0, "import_data_topkt: wfproc_translate_seedcodes() error.\n");
			return (-1);
		}
		if (ret == 0) {
			return (0);
		}
	} else {
		strcpy (tnet, data->net);
		strcpy (tsta, data->sta);
		strcpy (tchan, data->chan);
		strcpy (tloc, data->loc);
	}

	pkt = newPkt();
	if (pkt == NULL) {
		register_error (0, "import_data_topkt: newPkt() error.\n");
		return (-1);
	}
	pkt->pkttype = suffix2pkttype("GENC");
	pkt->nchannels = 1;
	achan = (PktChannel *) gettbl (pkt->channels, 0);
	if (wfproc_copyPktChannel(&achan, data) < 0) {
		register_error (0, "import_data_topkt: wfproc_copyPktChannel() error.\n");
		freePkt (pkt);
		return (-1);
	}
	strcpy (achan->net, tnet);
	strcpy (achan->sta, tsta);
	strcpy (achan->chan, tchan);
	strcpy (achan->loc, tloc);
	settbl (pkt->channels, 0, achan);

	settbl (pkts_tbl, -1, pkt);

	return (0);
}
