/* Copyright (c) 2008 Boulder Real Time Technologies, Inc. */
/* All rights reserved */
 
/* This software module is wholly owned by Boulder Real Time 
   Technologies, Inc. This software may be used freely in any 
   way as long as the copyright statement above is not removed. */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <strings.h>

#include "wfgather.h"
#include "brttutil.h"
#include "wffil.h"
#include "Pkt.h"
#include "tr.h"

#define SAMP(x,y)       (long)((x)<0.0?((x)*(y)-0.5):((x)*(y)+0.5))
#define INT(x)  (int)((x)<0.0?((x)-0.5):((x)+0.5)

int
wfgather_match_channel (void *wfgv, char *nscl)

{
	WFGather *wfg = (WFGather *) wfgv;
	int i;

	for (i=0; i<maxtbl(wfg->channels_accept); i++) {
		char *ptr;
		char line[512];
		char seed_expr[512];
		int ret;
		int j;

		ptr = (char *) gettbl (wfg->channels_accept, i);
		strcpy (line, ptr);
		ptr = strchr(line, '#');
		if (ptr) *ptr = '\0';
		strtrim(line);
		if (strlen(line) < 1) continue;

		ret = sscanf (line, "%s", seed_expr);
		if (ret < 1) {
			register_error (0, "wfgather_match_channel: Cannot parse channels_accept line %s.\n", line);
			return (-2);
		}
		if (strmatches (nscl, seed_expr, NULL) == 0) continue;
		if (wfg->channels_reject == NULL) return (i);
		for (j=0; j<maxtbl(wfg->channels_reject); j++) {
			ptr = (char *) gettbl (wfg->channels_reject, j);
			strcpy (line, ptr);
			ptr = strchr(line, '#');
			if (ptr) *ptr = '\0';
			strtrim(line);
			if (strlen(line) < 1) continue;

			ret = sscanf (line, "%s", seed_expr);
			if (ret < 1) {
				register_error (0, "wfgather_match_channel: Cannot parse channels_reject line %s.\n", line);
				return (-2);
			}
			if (strmatches (nscl, seed_expr, NULL) == 0) continue;
			return (-1);
		}
		return (i);
	}

	return (-1);
}

int
wfgather_process_dbmaster (void *wfgv)

{
	WFGather *wfg = (WFGather *) wfgv;
	Dbptr cal, site;
	char expr[128];
	long n;

	cal = dblookup (wfg->dbmaster, 0, "calibration", 0, 0);
	if (cal.table == dbINVALID) {
		register_error (0, "wfgather_process_dbmaster: Cannot lookup calibration table.\n");
		return (-1);
	}
	site = dblookup (wfg->dbmaster, 0, "site", 0, 0);
	if (site.table == dbINVALID) {
		register_error (0, "wfgather_process_dbmaster: Cannot lookup site table.\n");
		return (-1);
	}

	sprintf (expr, "%.3lf >= time && ( %.3lf <= endtime || endtime == null(\"endtime\") )", 
					wfg->time, wfg->time);
	cal = dbsubset (cal, expr, NULL);
	cal = dbjoin (cal, site, NULL, NULL, 0, NULL, NULL);

	dbquery (cal, dbRECORD_COUNT, &n);
	if (n < 1) {
		register_error (0, "wfgather_process_dbmaster: No calibration table rows to process.\n");
		return (-1);
	}

	if (wfg->channels_arr) {
		freearr (wfg->channels_arr, free);
		wfg->channels_arr = NULL;
	}
	if (wfg->channels_tbl) {
		freetbl (wfg->channels_tbl, NULL);
		wfg->channels_tbl = NULL;
	}
	if (wfg->gather) {
		free (wfg->gather);
		wfg->gather = NULL;
	}

	wfg->channels_arr = newarr (NULL);
	if (wfg->channels_arr == NULL) {
		register_error (0, "wfgather_process_dbmaster: newarr(channels_arr) error.\n");
		return (-1);
	}
	wfg->channels_tbl = newtbl (0);
	if (wfg->channels_tbl == NULL) {
		register_error (0, "wfgather_process_dbmaster: newtbl(channels_tbl) error.\n");
		return (-1);
	}
	wfg->samprate = 0.0;
	wfg->nchannels = 0;
	wfg->tstart = 0.0;
	for (cal.record=0; cal.record<n; cal.record++) {
		char sta[32], chan[32], refsta[32];
		char snet[32], ssta[32], schan[32], sloc[32];
		double samprate, calib, calper, lat, lon, elev, dnorth, deast;
		char segtype[32];
		char nscl[128];
		int ret;
		char *ptr;
		double clip_lower, clip_upper;
		char seed_expr[512];
		char line[512];
		char filter[512];
		size_t size;
		WFGatherChannel *wfgc;

		if (dbgetv (cal, 0, 	"sta", sta, 
					"chan", chan, 
					"refsta", refsta, 
					"samprate", &samprate,
					"calib", &calib,
					"calper", &calper,
					"segtype", segtype,
					"lat", &lat,
					"lon", &lon,
					"elev", &elev,
					"dnorth", &dnorth,
					"deast", &deast,
					NULL) < 0) {
			register_error (0, "wfgather_process_dbmaster: dbgetv(calibration) error.\n");
			return (-1);
		}
		if (dbseed_net (wfg->dbmaster, sta, snet, ssta) < 0) {
			/* complain (0, "wfgather_process_dbmaster: seed_net(%s) error - ignoring %s.\n", sta, sta); */
			continue;
		}
		if (dbseed_loc (wfg->dbmaster, sta, chan, schan, sloc) < 0) {
			/* complain (0, "wfgather_process_dbmaster: seed_loc(%s,%s) error - ignoring %s,%s.\n", sta, chan, sta, chan); */
			continue;
		}
		wfproc_make_nscl (snet, ssta, schan, sloc, nscl);
		ret = wfgather_match_channel (wfgv, nscl);
		if (ret == -2) {
			register_error (0, "wfgather_process_dbmaster: wfgather_match_channel(%s) error.\n", nscl);
			return (-1);
		}
		if (ret < 0) continue;

		ptr = (char *) gettbl (wfg->channels_accept, ret);
		clip_lower = -1.e30;
		clip_upper = 1.e30;
		strcpy (filter, "none");
		strcpy (line, ptr);
		ptr = strtok(line, " \t");
		while (1) {
			if (ptr == NULL) {
				ptr = (char *) gettbl (wfg->channels_accept, ret);
				register_error (0, "wfgather_process_dbmaster: Cannot parse seed_expr from '%s'.\n", ptr);
				return (-1);
			}
			if (strcmp(ptr, "")) break;
			ptr = strtok(NULL, " \t");
		}
		strcpy (seed_expr, ptr);
		while (1) {
			ptr = strtok(NULL, " \t");
			if (ptr == NULL) {
				goto OUT;
			}
			if (strcmp(ptr, "")) break;
		}
		if (strcmp(ptr, "-")) {
			clip_lower = atof(ptr);
		}
		while (1) {
			ptr = strtok(NULL, " \t");
			if (ptr == NULL) {
				goto OUT;
			}
			if (strcmp(ptr, "")) break;
		}
		if (strcmp(ptr, "-")) {
			clip_upper = atof(ptr);
		}
		while (1) {
			ptr = strtok(NULL, "");
			if (ptr == NULL) {
				goto OUT;
			}
			if (strcmp(ptr, "")) break;
		}
		strcpy (filter, ptr);
		strtrim (filter);
OUT:		if (wffil_parse_filter (NULL, filter) < 0) {
			register_error (0, "wfgather_process_dbmaster: wffil_parse_filter(%s) error for %s.\n", filter, nscl);
			return (-1);
		}

		ptr = (char *) getarr (wfg->channels_arr, nscl);
		if (ptr != NULL) {
			register_error (0, "wfgather_process_dbmaster: pre-existing channels entry for %s - probably a problem with the calibration table.\n", nscl);
			return (-1);
		}

		if (wfg->samprate == 0.0) {
			wfg->samprate = samprate;
		} else if (wfg->samprate != samprate) {
			complain (0, "wfgather_process_dbmaster: %s samprate of %.6lf different from %.6lf - cannot use this channel.\n", 
						nscl, samprate, wfg->samprate);
			continue;
		}

		size = sizeof(WFGatherChannel);
		wfgc = (WFGatherChannel *) malloc (size);
		if (wfgc == NULL) {
			register_error (1, "wfgather_process_dbmaster: malloc(WFGatherChannel,%ld) error for %s.\n", size, nscl);
			return (-1);
		}
		memset (wfgc, 0, size);

		strcpy (wfgc->snet, snet);
		strcpy (wfgc->ssta, ssta);
		strcpy (wfgc->schan, schan);
		strcpy (wfgc->sloc, sloc);
		strcpy (wfgc->sta, sta);
		strcpy (wfgc->chan, chan);
		strcpy (wfgc->filter, filter);
		strcpy (wfgc->segtype, segtype);
		wfgc->calib = calib;
		wfgc->calper = calper;
		wfgc->samprate = samprate;
		wfgc->lat = lat;
		wfgc->lon = lon;
		wfgc->elev = elev;
		wfgc->clip_upper = clip_upper;
		wfgc->clip_lower = clip_lower;
		wfgc->ichannel = wfg->nchannels;
		(wfg->nchannels)++;

		settbl (wfg->channels_tbl, -1, wfgc);
		setarr (wfg->channels_arr, nscl, wfgc);
		if (wfg->verbose) {
			WFProcTask *task;

			task = wfproc_task_get_from_obj (wfgv);
			elog_notify (0, "%s: Processing %s->%s %s\n", task->name, nscl, wfgc->sta, wfgc->chan);
		}
	}
	if (wfg->verbose) {
		WFProcTask *task;

		task = wfproc_task_get_from_obj (wfgv);
		elog_notify (0, "%s: Processing %d waveform channel\n", task->name, wfg->nchannels);
	}
	
	if (wfg->nchannels == 0) {
		register_error (0, "wfgather_process_dbmaster: No channels to process.\n");
		return (-1);
	}

	return (0);
}

void *
wfgather_new (Pf *params)

{
	char strbuf[512];
	int ret;
	size_t size;
	Dbptr dbmaster;
	double time;
	WFGather *wfg;
	WFProcDb *wfpdbm;

	if (parse_param (params, "dbmaster", P_STRCP, 1, strbuf) < 0) {
		register_error (0, "wfgather_new: parse_param(dbmaster) error.\n");
		return (NULL);
	}

	ret = sscanf (strbuf, "%lu", (unsigned long *) &wfpdbm);
	if (ret != 1) {
		register_error (0, "wfgather_new: Cannot parse 'dbmaster' parameter (%s).\n", strbuf);
		return (NULL);
	}

	if (wfprocdb_lock (wfpdbm, &dbmaster) < 0) {
		register_error (0, "wfgather_new: wfprocdb_lock() error.\n");
		return (NULL);
	}
	wfprocdb_unlock (wfpdbm);

	strcpy (strbuf, "");
	if (parse_param (params, "time", P_STRCP, 0, strbuf) < 0) {
		register_error (0, "wfgather_new: parse_param(time) error.\n");
		return (NULL);
	}
	if (!strcmp(strbuf, "")) {
		time = now();
	} else {
		if (zstr2epoch (strbuf, &time) < 0) {
			register_error (0, "wfgather_new: Cannot parse time value '%s'.\n", strbuf);
			return (NULL);
		}
	}

	size = sizeof(WFGather);
	wfg = (WFGather *) malloc (size);
	if (wfg == NULL) {
		register_error (1, "wfgather_new: malloc(WFGather,%ld) error.\n", size);
		return (NULL);
	}

	memset (wfg, 0, size);

	wfg->wfpdbm = wfpdbm;
	wfg->dbmaster = dbmaster;
	wfg->time = time;
	trfill_gap (&(wfg->gap_value), 1);
	wfg->clip_value = -wfg->gap_value;

	return (wfg);
}

int
wfgather_destroy (void *wfgv)

{
	if (wfgv == NULL) return (0);
	free (wfgv);

	return (0);
}

int
wfgather_config (void *wfgv, Pf *params, Pf *state)

{
	WFGather *wfg = (WFGather *) wfgv;
	Tbl *channels;
	Tbl *reject;
	double output_duration;
	double maximum_wait_time=0.0;

	if (parse_param (params, "channels", P_TBL, 1, &channels) < 0) {
		register_error (0, "wfgather_config: parse_param(channels) error.\n");
		return (-1);
	}
	if (parse_param (params, "output_duration", P_DBL, 1, &output_duration) < 0) {
		register_error (0, "wfgather_config: parse_param(output_duration) error.\n");
		return (-1);
	}
	if (parse_param (params, "maximum_wait_time", P_DBL, 0, &maximum_wait_time) < 0) {
		register_error (0, "wfgather_config: parse_param(maximum_wait_time) error.\n");
		return (-1);
	}
	reject = NULL;
	if (parse_param (params, "reject", P_TBL, 0, &reject) < 0) {
		register_error (0, "wfgather_config: parse_param(reject) error.\n");
		return (-1);
	}
	wfg->channels_accept = channels;
	wfg->channels_reject = reject;
	wfg->output_duration = output_duration;
	wfg->maximum_wait_time = maximum_wait_time;
	if (parse_param (params, "verbose", P_LINT, 0, &(wfg->verbose)) < 0) {
		register_error (0, "wfgather_config: parse_param(verbose) error.\n");
		return (-1);
	}
	if (parse_param (params, "start", P_STRCP, 0, &(wfg->start)) < 0) {
		register_error (0, "wfgather_config: parse_param(start) error.\n");
		return (-1);
	}

	if (wfgather_process_dbmaster(wfgv) < 0) {
		register_error (0, "wfgather_config: wfgather_process_dbmaster() error.\n");
		return (-1);
	}

	return (0);
}

int
wfgather_proc (void *wfgv, WFProcData *input, Tbl **output, Pf **state)

{
	WFGather *wfg = (WFGather *) wfgv;
	PktChannel *data;
	char nscl[128];
	WFGatherChannel *wfgc;
	double datatime, block_wall_time, dt, error, calib;
	int i, j, n, iblock;
	size_t size;

	*output = NULL;

	if (strcmp(input->type, "wf")) {
		register_error (0, "wfgather_proc: Wrong data type %s.\n", input->type);
		wfproc_data_free (input);
		return (-1);
	}

	data = (PktChannel *) input->data;

	wfproc_make_nscl (data->net, data->sta, data->chan, data->loc, nscl);

	wfgc = (WFGatherChannel *) getarr (wfg->channels_arr, nscl);

	if (wfgc == NULL) {
		wfproc_data_free (input);
		return (0);
	}

	error = 100.0*ABS(data->samprate-wfg->samprate)/wfg->samprate;
	if (error > 1.0) {
		register_error (0, "wfgather_proc: %s: samprate difference of %.lf%% too big.\n", nscl, error);
		wfproc_data_free (input);
		return (-1);
	}

	block_wall_time = now();
	if (wfg->gather == NULL) {
		int r;
		double shift;
		char *ptr;

		wfg->nsamps_per_tblock = (int)(wfg->output_duration * wfg->samprate + 0.5);
		if (wfg->nsamps_per_tblock < 1) wfg->nsamps_per_tblock = 1;
		wfg->nvalues_per_tblock = wfg->nsamps_per_tblock*wfg->nchannels;
		size = sizeof(float)*wfg->nvalues_per_tblock;
		wfg->gather = (float *) malloc (size);
		if (wfg->gather == NULL) {
			register_error (1, "wfgather_proc: %s: malloc(gather,%ld) error.\n", nscl, size);
			wfproc_data_free (input);
			return (-1);
		}
		wfg->ntblocks = 1;
		for (i=0; i<wfg->nvalues_per_tblock; i++) wfg->gather[i] = wfg->gap_value;
		ptr = wfg->start;
		r = 0;
		shift = 0.0;
		if (*ptr == 'r') {
			r = 1;
			ptr++;
		}
		if (strcmp(ptr, "")) {
			shift = atof(ptr);
		}
		wfg->tstart = data->time+shift;
		if (r) {
			double time_per_tblock;

			time_per_tblock = wfg->nsamps_per_tblock/wfg->samprate;
			r = (int)(wfg->tstart/time_per_tblock);
			r++;
			wfg->tstart = r*time_per_tblock;
		}
		wfg->block_wall_times_nentries = 1;
		size = sizeof(double)*wfg->block_wall_times_nentries;
		wfg->block_wall_times = (double *) malloc (size);
		if (wfg->block_wall_times == NULL) {
			register_error (1, "wfgather_proc: %s: malloc(gather->block_wall_times,%ld) error.\n", nscl, size);
			wfproc_data_free (input);
			return (-1);
		}
		for (j=0; j<wfg->block_wall_times_nentries; j++) wfg->block_wall_times[j] = block_wall_time;
	}

	dt = 1.0/data->samprate;
	datatime = data->time;
	wfg->nvalues = wfg->nvalues_per_tblock*wfg->ntblocks;
	wfg->nsamps = wfg->nsamps_per_tblock*wfg->ntblocks;
	calib = data->calib;
	if (!strcmp(data->segtype, "c")) calib = 0.0;
	if (calib == 0.0) {
		calib = wfgc->calib;
		if (!strcmp(wfgc->segtype, "c")) calib = 0.0;
	}
	if (calib == 0.0) calib = 1.0;
	for (i=0; i<data->nsamp; i++) {
		float value;

		j = SAMP((data->time+i*dt-wfg->tstart),data->samprate);
		if (j < 0) continue;
		while (j >= wfg->nsamps) {
			int k;

			n = (int)((double)(j - wfg->nsamps + 1)/wfg->nsamps_per_tblock);
			n++;
			size = sizeof(float)*wfg->nvalues_per_tblock*(wfg->ntblocks+n);
			wfg->gather = (float *) realloc (wfg->gather, size);
			if (wfg->gather == NULL) {
				register_error (1, "wfgather_proc: %s: realloc(gather,%ld) error.\n", nscl, size);
				wfproc_data_free (input);
				return (-1);
			}
			wfg->ntblocks += n;
			wfg->nsamps += n*wfg->nsamps_per_tblock;
			n *= wfg->nvalues_per_tblock;
			for (k=wfg->nvalues; k<wfg->nvalues+n; k++) wfg->gather[k] = wfg->gap_value;
			wfg->nvalues += n;
		}
		if (data->isfloat) {
			memcpy (&value, &(data->data[i]), sizeof(float));
		} else {
			if (data->data[i] == BADS4) {
				value = wfg->gap_value;
			} else {
				value = data->data[i];
			}
		}
		if (value != wfg->gap_value) {
			if (value < wfgc->clip_lower || value > wfgc->clip_upper) value = wfg->clip_value;
			else value *= (float)calib;
		}
		wfg->gather[j*wfg->nchannels+wfgc->ichannel] = value;
		iblock = j/wfg->nsamps_per_tblock;
		if (iblock >= wfg->block_wall_times_nentries) {
			n = iblock+1;
			size = sizeof(double)*n;
			wfg->block_wall_times = (double *) realloc (wfg->block_wall_times, size);
			if (wfg->block_wall_times == NULL) {
				register_error (1, "wfgather_proc: %s: realloc(gather->block_wall_times,%ld) error.\n", nscl, size);
				wfproc_data_free (input);
				return (-1);
			}
			size = sizeof(double)*(n-wfg->block_wall_times_nentries);
			for (j=wfg->block_wall_times_nentries; j<n; j++) wfg->block_wall_times[j] = block_wall_time;
			wfg->block_wall_times_nentries = n;
		}
		wfg->block_wall_times[iblock] = block_wall_time;
	}
	wfproc_data_free (input);

	for (iblock=0,i=0; iblock<wfg->ntblocks; iblock++) {
		int ok;
		WFProcData *output_packet;
		double outtime;

		ok = 1;
		for (j=0; j<wfg->nvalues_per_tblock; j++,i++) {
			if (wfg->gather[i] == wfg->gap_value) ok = 0;
		} 
		outtime = wfg->tstart + iblock*wfg->nsamps_per_tblock/wfg->samprate;
		if (ok == 1 || block_wall_time-wfg->block_wall_times[iblock] > wfg->maximum_wait_time
						|| datatime-outtime > wfg->maximum_wait_time) {
			output_packet = wfgather_create_output (wfg, iblock);
			if (output_packet == NULL) {
				register_error (0, "wfgather_proc: %s: wfgather_create_output() error.\n", nscl);
				return (-1);
			}
			if (*output == NULL) {
				*output = newtbl(1);
				if (*output == NULL) {
					register_error (0, "wfgather_proc: %s: newtbl(output) error.\n", nscl);
					return (-1);
				}
			}
			settbl (*output, -1, output_packet);
		} else break;
	}

	if (iblock > 0) {
		float *gather;
		size_t size2;
		double *block_wall_times;

		n = wfg->ntblocks-iblock;
		if (n < 1) {
			size = sizeof(float)*wfg->nvalues_per_tblock;
			size2 = sizeof(double);
			wfg->ntblocks = 1;
			wfg->block_wall_times_nentries = 1;
		} else {
			size = sizeof(float)*n*wfg->nvalues_per_tblock;
			size2 = sizeof(double)*n;
			wfg->ntblocks = n;
			wfg->block_wall_times_nentries = n;
		}
		gather = (float *) malloc (size);
		if (gather == NULL) {
			register_error (1, "wfgather_proc: %s: malloc(newgather,%ld) error.\n", nscl, size);
			return (-1);
		}
		block_wall_times = (double *) malloc (size2);
		if (block_wall_times == NULL) {
			register_error (1, "wfgather_proc: %s: malloc(newblock_wall_times,%ld) error.\n", nscl, size2);
			free (gather);
			return (-1);
		}
		wfg->nvalues = wfg->nvalues_per_tblock*wfg->ntblocks;
		wfg->nsamps = wfg->nsamps_per_tblock*wfg->ntblocks;
		wfg->tstart += iblock*wfg->nsamps_per_tblock*dt;
		for (i=0; i<wfg->nvalues; i++) gather[i] = wfg->gap_value;
		for (i=0; i<wfg->block_wall_times_nentries; i++) block_wall_times[i] = block_wall_time;
		size = sizeof(float)*n*wfg->nvalues_per_tblock;
		size2 = sizeof(double)*n;
		if (size > 0) {
			memcpy (gather, &(wfg->gather[iblock*wfg->nvalues_per_tblock]), size); 
			memcpy (block_wall_times, &(wfg->block_wall_times[iblock]), size2); 
		}
		free (wfg->gather);
		wfg->gather = gather;
		free (wfg->block_wall_times);
		wfg->block_wall_times = block_wall_times;
	}

	return (0);
}

WFProcData *
wfgather_create_output (WFGather *wfg, int iblock)

{
	WFProcData *data=NULL;
	WFGatherOutput *out;
	int i, j, k, ioff;
	size_t size, out_size;
	double dt;

	size = sizeof(WFGatherOutput);
	out_size = size;
	out = (WFGatherOutput *) malloc (size);
	if (out == NULL) {
		register_error (1, "wfgather_create_output: malloc(WFGatherOutput,%ld) error.\n", size);
		return (NULL);
	}
	memset (out, 0, size);
	out->channels = wfg->channels_tbl;
	out->tstart = wfg->tstart + iblock*wfg->nsamps_per_tblock/wfg->samprate;
	out->samprate = wfg->samprate;
	out->nchannels = wfg->nchannels;
	out->nsamps = wfg->nsamps_per_tblock;
	out->nvalues = out->nchannels*out->nsamps;
	out->clip_value = wfg->clip_value;
	out->gap_value = wfg->gap_value;
	size = sizeof(float)*wfg->nvalues_per_tblock;
	out_size += size;
	out->gather = (float *) malloc (size);
	if (out->gather == NULL) {
		register_error (1, "wfgather_create_output: malloc(WFGatherOutput->gather,%ld) error.\n", size);
		free (out);
		return (NULL);
	}
	ioff = iblock*wfg->nvalues_per_tblock;
	dt = 1.0/out->samprate;
	for (i=0,k=0; i<out->nchannels; i++) {
		WFGatherChannel *chan;
		int k0;

		chan = (WFGatherChannel *) gettbl (wfg->channels_tbl, i);
		k0 = k;
		if (!strcmp(chan->filter, "none")) {
			for (j=0; j<out->nsamps; j++,k++) {
				out->gather[k] = wfg->gather[ioff+j*wfg->nchannels+chan->ichannel];
			}
		} else {
			int nsamps;
			double tstart, dt2;
			float *dptr;

			for (j=0; j<out->nsamps; j++,k++) {
				out->gather[k] = wfg->gather[ioff+j*wfg->nchannels+chan->ichannel];
				if (out->gather[k] == out->clip_value) out->gather[k] = out->gap_value;
			}
			dt2 = dt;
			nsamps = out->nsamps;
			tstart = out->tstart;
			dptr = &(out->gather[k0]);
			if (wffil (NULL, chan->filter, (double) 0.5, &nsamps,
						&tstart, &dt2, &dptr, &nsamps, NULL, NULL, 
						&(chan->filter_hook)) < 0) {
				register_error (0, "wfgather_create_output: wffil() error.\n");
				free (out->gather);
				free (out);
				return (NULL);
			}
			for (j=0,k=k0; j<out->nsamps; j++,k++) {
				if (wfg->gather[ioff+j*wfg->nchannels+chan->ichannel] == out->clip_value) {
					out->gather[k] = out->clip_value;
				}
			}
		}
	}

	for (i=0; i<out->nvalues; i++) {
		if (out->gather[i] == out->gap_value) (out->ngaps)++;
		if (out->gather[i] == out->clip_value) (out->nclips)++;
	}
	
	data = wfproc_data_new ("wfgather");
	if (data == NULL) {
		register_error (0, "wfgather_create_output: wfproc_data_new() error.\n");
		free (out->gather);
		free (out);
		return (NULL);
	}

	data->data = out;
	data->printproc = wfgather_data_print; 
	data->freeproc = wfgather_data_free;
	data->copyproc = wfgather_data_copy;
	data->topktproc = wfgather_data_topkt;
	data->size = sizeof(WFProcData) + out_size;

	return (data);
}

char *
wfgather_data_print (WFProcData *data, int verbose)

{
	char line[256];
	char *s;
	WFGatherOutput *out;

	out = (WFGatherOutput *) data->data;

	s = strtime(out->tstart);

	sprintf (line, "%s nchans=%d nsamps= %d ngaps=%d, nclips=%d", s, out->nchannels, out->nsamps, out->ngaps, out->nclips);

	free (s);

	return (strdup(line));
}

void
wfgather_data_free (WFProcData *data)

{
	WFGatherOutput *out;

	if (data == NULL) return;
	out = (WFGatherOutput *) data->data;
	if (out) {
		if (out->gather) free (out->gather);
		free (out);
	}
	free (data);
}

WFProcData *
wfgather_data_copy (WFProcData *data)

{
	WFProcData *copy=NULL;
	WFGatherOutput *out, *copyout;
	size_t size;

	copy = wfproc_data_new ("wfgather");
	if (copy == NULL) {
		register_error (0, "wfgather_data_copy: wfproc_data_new() error.\n");
		return (NULL);
	}

	*copy = *data;

	size = sizeof(WFGatherOutput);
	copyout = (WFGatherOutput *) malloc(size);
	if (copyout == NULL) {
		register_error (1, "wfgather_data_copy: malloc(WFGatherOutput,%ld) error.\n", size);
		free (copy);
		return (NULL);
	}

	out = (WFGatherOutput *) data->data;
	*copyout = *out;
	copy->data = copyout;

	size = sizeof(float)*out->nvalues;
	copyout->gather = (float *) malloc (size);
	if (copyout->gather == NULL) {
		register_error (1, "wfgather_data_copy: malloc(copyout->gather,%ld) error.\n", size);
		free (copyout);
		free (copy);
		return (NULL);
	}
	memcpy (copyout->gather, out->gather, size);

	return (copy);
}

int
wfgather_data_topkt (WFProcData *input, Tbl *translations, Tbl *pkts_tbl)

{
	WFGatherOutput *data;
	int i;
	Packet *pkt;

	data = (WFGatherOutput *) input->data;
	pkt = newPkt();
	if (pkt == NULL) {
		register_error (0, "wfgather_data_topkt: newPkt() error.\n");
		return (-1);
	}
	pkt->pkttype = suffix2pkttype("flt");

	pkt->nchannels = 0;
	for (i=0; i<maxtbl(data->channels); i++) {
		WFGatherChannel *chan;
		size_t size;
		int j, ret;
		char tnet[32], tsta[32], tchan[32], tloc[32];
		PktChannel *achan, achan_tmp;

		chan = (WFGatherChannel *) gettbl (data->channels, i);
		if (translations) {
			ret = wfproc_translate_seedcodes (chan->snet, chan->ssta, chan->schan, chan->sloc,
							translations, tnet, tsta, tchan, tloc);
			if (ret < 0) {
				register_error (0, "wfgather_data_topkt: wfproc_translate_seedcodes() error.\n");
				freePkt (pkt);
				return (-1);
			}
			if (ret == 0) {
				continue;
			}
		} else {
			strcpy (tnet, chan->snet);
			strcpy (tsta, chan->ssta);
			strcpy (tchan, chan->schan);
			strcpy (tloc, chan->sloc);
		}
		memset (&achan_tmp, 0, sizeof(PktChannel));
		strcpy (achan_tmp.net, tnet);
		strcpy (achan_tmp.sta, tsta);
		strcpy (achan_tmp.chan, tchan);
		strcpy (achan_tmp.loc, tloc);
		strcpy (achan_tmp.segtype, chan->segtype);
		achan_tmp.calib = chan->calib;
		achan_tmp.calper = chan->calper;
		achan_tmp.time = data->tstart;
		achan_tmp.nsamp = data->nsamps;
		achan_tmp.datasz = data->nsamps;
		achan_tmp.samprate = data->samprate;
		achan_tmp.isfloat = 1;
		size = achan_tmp.datasz*sizeof(int);
		achan_tmp.data = (int *) malloc(size);
		if (achan_tmp.data == NULL) {
			register_error (1, "wfgather_data_topkt: malloc(achan->data,%ld) error.\n", size);
			freePkt (pkt);
			return (-1);
		}
		for (j=0; j<achan_tmp.nsamp; j++) {
			float fvalue;

			fvalue = data->gather[j+chan->ichannel*data->nsamps];
			if (fvalue == data->clip_value) {
				fvalue = data->gap_value;
			}
			if (fvalue != data->gap_value) {
				if (chan->calib != 0.0) {
					fvalue /= (float)chan->calib;
				}
			}
			memcpy (&(achan_tmp.data[j]), &fvalue, 4);
		}
		achan = (PktChannel *) gettbl (pkt->channels, pkt->nchannels);
		if (wfproc_copyPktChannel(&achan, &achan_tmp) < 0) {
			register_error (0, "wfgather_data_topkt: wfproc_copyPktChannel() error.\n");
			freePkt (pkt);
			return (-1);
		}
		free (achan_tmp.data);
		settbl (pkt->channels, pkt->nchannels, achan);
		(pkt->nchannels)++;
	}

	for (i=pkt->nchannels; i<maxtbl(pkt->channels); i++) {
		PktChannel *achan;

		achan = (PktChannel *) deltbl (pkt->channels, i);
		if (achan) freePktChannel (achan);
	}

	if (pkt->nchannels == 0) {
		freePkt (pkt);
		return (0);
	}

	settbl (pkts_tbl, -1, pkt);

	return (0);
}

Tbl *
wfgather_channels_copy (Tbl *channels)

{
	Tbl *copy;
	size_t size;
	int i;

	if (channels == NULL) return (NULL);

	copy = newtbl (0);
	if (copy == NULL) {
		register_error (0, "wfgather_channels_copy: newtbl(copy) error.\n");
		return (NULL);
	}

	size = sizeof(WFGatherChannel);
	for (i=0; i<maxtbl(channels); i++) {
		WFGatherChannel *chan, *newchan;

		chan = (WFGatherChannel *) gettbl (channels, i);

		newchan = (WFGatherChannel *) malloc (size);
		if (newchan == NULL) {
			register_error (1, "wfgather_channels_copy: malloc(WFGatherChannel,%ld) error.\n", size);
			freetbl (copy, free);
			return (NULL);
		}
		*newchan = *chan;
		settbl (copy, -1, newchan);
	}

	return (copy);
}

int
wfgather_channels_match (Tbl *channels1, Tbl *channels2)

{
	int i, n;

	n = maxtbl(channels2) ;

	for (i=0; i<maxtbl(channels1); i++) {
		WFGatherChannel *chan1, *chan2;

		chan1 = (WFGatherChannel *) gettbl (channels1, i);
		if (i >= n) return (0);
		chan2 = (WFGatherChannel *) gettbl (channels2, i);
		if (strcmp(chan1->chan, chan2->chan)) return (0);
		if (strcmp(chan1->sta, chan2->sta)) return (0);
		if (strcmp(chan1->filter, chan2->filter)) return (0);
	}

	if (i != n) return (0);

	return (1);
}


