/* Copyright (c) 2008 Boulder Real Time Technologies, Inc. */
/* All rights reserved */
 
/* This software module is wholly owned by Boulder Real Time 
   Technologies, Inc. This software may be used freely in any 
   way as long as the copyright statement above is not removed. */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>

#include "wfstack.h"

extern int verbose;

void *
wfstack_new (Pf *params)

{
	char strbuf[512];
	int ret;
	size_t size;
	Dbptr dbmaster;
	double time;
	WFStack *wfs;
	WFProcDb *wfpdbm;

	if (parse_param (params, "dbmaster", P_STRCP, 1, strbuf) < 0) {
		register_error (0, "wfstack_new: parse_param(dbmaster) error.\n");
		return (NULL);
	}

	ret = sscanf (strbuf, "%lu", (unsigned long *) &wfpdbm);
	if (ret != 1) {
		register_error (0, "wfstack_new: Cannot parse 'dbmaster' parameter (%s).\n", strbuf);
		return (NULL);
	}

	if (wfprocdb_lock (wfpdbm, &dbmaster) < 0) {
		register_error (0, "wfstack_new: wfprocdb_lock() error.\n");
		return (NULL);
	}
	wfprocdb_unlock(wfpdbm);

	strcpy (strbuf, "");
	if (parse_param (params, "time", P_STRCP, 0, strbuf) < 0) {
		register_error (0, "wfstack_new: parse_param(time) error.\n");
		return (NULL);
	}
	if (!strcmp(strbuf, "")) {
		time = now();
	} else {
		if (zstr2epoch (strbuf, &time) < 0) {
			register_error (0, "wfstack_new: Cannot parse time value '%s'.\n", strbuf);
			return (NULL);
		}
	}

	size = sizeof(WFStack);
	wfs = (WFStack *) malloc (size);
	if (wfs == NULL) {
		register_error (1, "wfstack_new: malloc(WFStack,%ld) error.\n", size);
		return (NULL);
	}

	memset (wfs, 0, size);

	wfs->wfpdbm = wfpdbm;
	wfs->dbmaster = dbmaster;
	wfs->time = time;

	return (wfs);
}

int
wfstack_destroy (void *wfsv)

{
	WFStack *wfs = (WFStack *) wfsv;

	if (wfsv == NULL) return (0);

	if (wfs->stack) stack_destroy (wfs->stack);

	free (wfsv);

	return (0);
}

int
wfstack_config (void *wfsv, Pf *params, Pf *state)

{
	WFStack *wfs = (WFStack *) wfsv;

	wfs->stack = stack_new (params, verbose);
	if (wfs->stack == NULL) {
		register_error (0, "wfstack_config: stack_new() error.\n");
		return (-1);
	}

	return (0);
}

int
wfstack_proc (void *wfsv, WFProcData *input, Tbl **output, Pf **state)

{
	WFStack *wfs = (WFStack *) wfsv;

	if (stack_proc (wfs->stack, input, output, state) < 0) {
		register_error (0, "wfstack_proc: stack_proc() error.\n");
		return (-1);
	}

	return (0);
}

WFProcData *
wfstack_create_output (WFStack *wfs, int nsamps, char *output_variable, Tbl *pktoutput_channels, WFGatherOutput *gdata)

{
	WFProcData *data=NULL;

	data = stack_create_output (wfs->stack, nsamps, output_variable, pktoutput_channels, gdata);

	return (data);
}

char *
wfstack_data_print (WFProcData *data, int verbose)

{
	char *out=NULL;

	out = stack_data_print (data, verbose);

	return (out);
}

void
wfstack_data_free (WFProcData *data)

{
	stack_data_free (data);
}

WFProcData *
wfstack_data_copy (WFProcData *data)

{
	WFProcData *copy=NULL;

	copy = stack_data_copy (data);

	return (copy);
}

int
wfstack_data_topkt (WFProcData *input, Tbl *translations, Tbl *pkts_tbl)

{
	return (stack_data_topkt (input, translations, pkts_tbl));
}
