/* Copyright (c) 2008 Boulder Real Time Technologies, Inc. */
/* All rights reserved */
 
/* This software module is wholly owned by Boulder Real Time 
   Technologies, Inc. This software may be used freely in any 
   way as long as the copyright statement above is not removed. */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>

#include "wftest.h"
#include "import.h"
#include "Pkt.h"
#include "coords.h"

#define INT(x)  (int)((x)<0.0?((x)-0.5):((x)+0.5))

int
wftest_match_channel (void *wftv, char *nscl)

{
	WFTest *wft = (WFTest *) wftv;
	int i;

	for (i=0; i<maxtbl(wft->channels_accept); i++) {
		char *ptr;
		char line[512];
		char seed_expr[512];
		int ret;
		int j;

		ptr = (char *) gettbl (wft->channels_accept, i);
		strcpy (line, ptr);
		ptr = strchr(line, '#');
		if (ptr) *ptr = '\0';
		strtrim(line);
		if (strlen(line) < 1) continue;

		ret = sscanf (line, "%s", seed_expr);
		if (ret < 1) {
			register_error (0, "wftest_match_channel: Cannot parse channels_accept line %s.\n", line);
			return (-2);
		}
		if (strmatches (nscl, seed_expr, NULL) == 0) continue;
		if (wft->channels_reject == NULL) return (i);
		for (j=0; j<maxtbl(wft->channels_reject); j++) {
			ptr = (char *) gettbl (wft->channels_reject, j);
			strcpy (line, ptr);
			ptr = strchr(line, '#');
			if (ptr) *ptr = '\0';
			strtrim(line);
			if (strlen(line) < 1) continue;

			ret = sscanf (line, "%s", seed_expr);
			if (ret < 1) {
				register_error (0, "wftest_match_channel: Cannot parse channels_reject line %s.\n", line);
				return (-2);
			}
			if (strmatches (nscl, seed_expr, NULL) == 0) continue;
			return (-1);
		}
		return (i);
	}

	return (-1);
}

int
wftest_process_dbmaster (void *wftv)

{
	WFTest *wft = (WFTest *) wftv;
	Dbptr cal, site;
	char expr[128];
	long n;

	cal = dblookup (wft->dbmaster, 0, "calibration", 0, 0);
	if (cal.table == dbINVALID) {
		register_error (0, "wftest_process_dbmaster: Cannot lookup calibration table.\n");
		return (-1);
	}
	site = dblookup (wft->dbmaster, 0, "site", 0, 0);
	if (site.table == dbINVALID) {
		register_error (0, "wftest_process_dbmaster: Cannot lookup site table.\n");
		return (-1);
	}

	sprintf (expr, "%.3lf >= time && ( %.3lf <= endtime || endtime == null(\"endtime\") )", 
					wft->dbtime, wft->dbtime);
	cal = dbsubset (cal, expr, NULL);
	cal = dbjoin (cal, site, NULL, NULL, 0, NULL, NULL);

	dbquery (cal, dbRECORD_COUNT, &n);
	if (n < 1) {
		register_error (0, "wftest_process_dbmaster: No calibration table rows to process.\n");
		return (-1);
	}

	if (wft->channels_arr) {
		freearr (wft->channels_arr, free);
		wft->channels_arr = NULL;
	}
	if (wft->channels_tbl) {
		freetbl (wft->channels_tbl, NULL);
		wft->channels_tbl = NULL;
	}

	wft->channels_arr = newarr (NULL);
	if (wft->channels_arr == NULL) {
		register_error (0, "wftest_process_dbmaster: newarr(channels_arr) error.\n");
		return (-1);
	}
	wft->channels_tbl = newtbl (0);
	if (wft->channels_tbl == NULL) {
		register_error (0, "wftest_process_dbmaster: newtbl(channels_tbl) error.\n");
		return (-1);
	}
	wft->samprate = 0.0;
	for (cal.record=0; cal.record<n; cal.record++) {
		char sta[32], chan[32], refsta[32];
		char snet[32], ssta[32], schan[32], sloc[32];
		double samprate, calib, calper, lat, lon, elev, dnorth, deast;
		char segtype[32];
		char nscl[128];
		int ret;
		char *ptr;
		double amp, toffset;
		char seed_expr[512];
		char line[512];
		size_t size;
		WFTestChannel *wftc;

		if (dbgetv (cal, 0, 	"sta", sta, 
					"chan", chan, 
					"refsta", refsta, 
					"samprate", &samprate,
					"calib", &calib,
					"calper", &calper,
					"segtype", segtype,
					"lat", &lat,
					"lon", &lon,
					"elev", &elev,
					"dnorth", &dnorth,
					"deast", &deast,
					NULL) < 0) {
			register_error (0, "wftest_process_dbmaster: dbgetv(calibration) error.\n");
			return (-1);
		}
		dbseed_net (wft->dbmaster, sta, snet, ssta);
		dbseed_loc (wft->dbmaster, sta, chan, schan, sloc);
		clear_register (0);
		wfproc_make_nscl (snet, ssta, schan, sloc, nscl);
		ret = wftest_match_channel (wftv, nscl);
		if (ret == -2) {
			register_error (0, "wftest_process_dbmaster: wftest_match_channel(%s) error.\n", nscl);
			return (-1);
		}
		if (ret < 0) continue;

		ptr = (char *) gettbl (wft->channels_accept, ret);
		strcpy (line, ptr);
		ptr = strtok(line, " \t");
		while (1) {
			if (ptr == NULL) {
				ptr = (char *) gettbl (wft->channels_accept, ret);
				register_error (0, "wftest_process_dbmaster: Cannot parse seed_expr from '%s'.\n", ptr);
				return (-1);
			}
			if (strcmp(ptr, "")) break;
			ptr = strtok(NULL, " \t");
		}
		strcpy (seed_expr, ptr);
		while (1) {
			ptr = strtok(NULL, " \t");
			if (ptr == NULL) {
				ptr = (char *) gettbl (wft->channels_accept, ret);
				register_error (0, "wftest_process_dbmaster: Cannot parse amp from '%s'.\n", ptr);
				return (-1);
			}
			if (strcmp(ptr, "")) break;
		}
		amp = atof(ptr);
		toffset = 0.0;
		while (1) {
			ptr = strtok(NULL, "");
			if (ptr == NULL) {
				goto OUT;
			}
			if (strcmp(ptr, "")) break;
		}
		toffset = atof(ptr);

OUT:		ptr = (char *) getarr (wft->channels_arr, nscl);
		if (ptr != NULL) {
			register_error (0, "wftest_process_dbmaster: pre-existing channels entry for %s - probably a problem with the calibration table.\n", nscl);
			return (-1);
		}

		if (wft->samprate == 0.0) {
			wft->samprate = samprate;
		} else if (wft->samprate != samprate) {
			complain (0, "wftest_process_dbmaster: %s samprate of %.6lf different from %.6lf - cannot use this channel.\n", 
						nscl, samprate, wft->samprate);
			continue;
		}

		size = sizeof(WFTestChannel);
		wftc = (WFTestChannel *) malloc (size);
		if (wftc == NULL) {
			register_error (1, "wftest_process_dbmaster: malloc(WFTestChannel,%ld) error for %s.\n", size, nscl);
			return (-1);
		}
		memset (wftc, 0, size);

		strcpy (wftc->snet, snet);
		strcpy (wftc->ssta, ssta);
		strcpy (wftc->schan, schan);
		strcpy (wftc->sloc, sloc);
		strcpy (wftc->sta, sta);
		strcpy (wftc->chan, chan);
		strcpy (wftc->segtype, segtype);
		wftc->calib = calib;
		wftc->calper = calper;
		wftc->samprate = samprate;
		wftc->lat = lat;
		wftc->lon = lon;
		wftc->elev = elev;
		wftc->amplitude = amp;
		wftc->tshift = toffset;
		wftc->achan = newPktChannel();
		if (wftc->achan == NULL) {
			register_error (0, "wftest_process_dbmaster: newPktChannel(wftc->achan) error.\n");
			return (-1);
		}
		strcpy (wftc->achan->net, snet);
		strcpy (wftc->achan->sta, ssta);
		strcpy (wftc->achan->chan, schan);
		strcpy (wftc->achan->loc, sloc);
		strcpy (wftc->achan->segtype, segtype);
		wftc->achan->calib = calib;
		wftc->achan->calper = calper;

		if (wft->latr == wftc->lat && wft->lonr == wftc->lon) {
			wftc->delx = 0.0;
			wftc->dely = 0.0;
		} else {
			double delta, azimuth;

			dist (wft->latr*M_PI/180.0, wft->lonr*M_PI/180.0,
					wftc->lat*M_PI/180.0, wftc->lon*M_PI/180.0, &delta, &azimuth);
			delta = deg2km (delta*180.0/M_PI);
			wftc->delx = delta*sin(azimuth);
			wftc->dely = delta*cos(azimuth);
		}

		settbl (wft->channels_tbl, -1, wftc);
		setarr (wft->channels_arr, nscl, wftc);
		(wft->nchannels)++;
		if (wft->verbose) {
			WFProcTask *task;

			task = wfproc_task_get_from_obj (wftv);
			elog_notify (0, "%s: Processing %s->%s %s\n", task->name, nscl, wftc->sta, wftc->chan);
		}
	}
	
	if (wft->nchannels == 0) {
		register_error (0, "wftest_process_dbmaster: No channels to process.\n");
		return (-1);
	}

	return (0);
}

void *
wftest_new (Pf *params)

{
	char strbuf[512];
	int ret;
	Dbptr dbmaster;
	double time;
	size_t size;
	WFTest *wft;
	WFProcDb *wfpdbm;

	if (parse_param (params, "dbmaster", P_STRCP, 1, strbuf) < 0) {
		register_error (0, "wftest_new: parse_param(dbmaster) error.\n");
		return (NULL);
	}

	ret = sscanf (strbuf, "%lu", (unsigned long *) &wfpdbm);
	if (ret != 1) {
		register_error (0, "wftest_new: Cannot parse 'dbmaster' parameter (%s).\n", strbuf);
		return (NULL);
	}

	if (wfprocdb_lock (wfpdbm, &dbmaster) < 0) {
		register_error (0, "wftest_new: wfprocdb_lock() error.\n");
		return (NULL);
	}
	wfprocdb_unlock (wfpdbm);

	strcpy (strbuf, "");
	if (parse_param (params, "time", P_STRCP, 0, strbuf) < 0) {
		register_error (0, "wftest_new: parse_param(time) error.\n");
		return (NULL);
	}
	if (!strcmp(strbuf, "")) {
		time = now();
	} else {
		if (zstr2epoch (strbuf, &time) < 0) {
			register_error (0, "wftest_new: Cannot parse time value '%s'.\n", strbuf);
			return (NULL);
		}
	}

	size = sizeof(WFTest);
	wft = (WFTest *) malloc (size);
	if (wft == NULL) {
		register_error (1, "wftest_new: malloc(WFTest,%ld) error.\n", size);
		return (NULL);
	}

	memset (wft, 0, size);
	wft->wfpdbm = wfpdbm;
	wft->dbmaster = dbmaster;
	wft->dbtime = time;

	wft->channels_tbl = newtbl(0);
	if (wft->channels_tbl == NULL) {
		register_error (0, "wftest_new: newtbl() error.\n");
		free (wft);
		return (NULL);
	}

	return (wft);
}

int
wftest_destroy (void *wftv)

{
	WFTest *wft = (WFTest *) wftv;

	if (wftv == NULL) return (0);

	if (wft->channels_tbl) freetbl (wft->channels_tbl, free);

	free (wftv);

	return (0);
}

int
wftest_config (void *wftv, Pf *params, Pf *state)

{
	WFTest *wft = (WFTest *) wftv;
	int yd;
	char *ptr;
	char snet[32], ssta[32];
	char expr[512];
	Dbptr site;
	long n;

	if (parse_param (params, "signal_type", P_STRCP, 1, wft->signal_type) < 0) {
		register_error (0, "wftest_config: parse_param(signal_type) error.\n");
		return (-1);
	}
	if (parse_param (params, "nsamps", P_LINT, 1, &(wft->nsamps)) < 0) {
		register_error (0, "wftest_config: parse_param(nsamps) error.\n");
		return (-1);
	}
	if (parse_param (params, "start", P_STRCP, 0, wft->start) < 0) {
		register_error (0, "wftest_config: parse_param(start) error.\n");
		return (-1);
	}
	if (parse_param (params, "duration", P_DBL, 0, &(wft->duration)) < 0) {
		register_error (0, "wftest_config: parse_param(duration) error.\n");
		return (-1);
	}
	if (parse_param (params, "repeat", P_DBL, 0, &(wft->repeat)) < 0) {
		register_error (0, "wftest_config: parse_param(repeat) error.\n");
		return (-1);
	}
	if (wft->duration > 0.0 && wft->repeat > 0.0 && wft->repeat <= wft->duration) {
		register_error (0, "wftest_config: repeat parameter must be greater than duration parameter.\n");
		return (-1);
	}

	if (!strcmp(wft->signal_type, "random")) {
	} else if (!strcmp(wft->signal_type, "sine")) {
		if (parse_param (params, "frequency", P_DBL, 1, &(wft->param1)) < 0) {
			register_error (0, "wftest_config: parse_param(frequency) error.\n");
			return (-1);
		}
		wft->param1 *= 2.0*M_PI;
	} else if (!strcmp(wft->signal_type, "dampedsine")) {
		if (parse_param (params, "frequency", P_DBL, 1, &(wft->param1)) < 0) {
			register_error (0, "wftest_config: parse_param(frequency) error.\n");
			return (-1);
		}
		wft->param1 *= 2.0*M_PI;
		if (parse_param (params, "damping", P_DBL, 1, &(wft->param2)) < 0) {
			register_error (0, "wftest_config: parse_param(damping) error.\n");
			return (-1);
		}
	} else {
		register_error (0, "wftest_config: Unrecognized signal_type '%s'.\n", wft->signal_type);
		return (-1);
	}

	if (parse_param (params, "channels", P_TBL, 1, &(wft->channels_accept)) < 0) {
		register_error (0, "wftest_config: parse_param(channels) error.\n");
		return (-1);
	}
	if (parse_param (params, "reject", P_TBL, 0, &(wft->channels_reject)) < 0) {
		register_error (0, "wftest_config: parse_param(reject) error.\n");
		return (-1);
	}
	if (parse_param (params, "verbose", P_LINT, 0, &(wft->verbose)) < 0) {
		register_error (0, "wftest_config: parse_param(verbose) error.\n");
		return (-1);
	}

	if (parse_param (params, "refsta", P_STRCP, 1, wft->refsta_seed) < 0) {
		register_error (0, "wftest_config: parse_param(refsta) error.\n");
		return (-1);
	}
	ptr = strchr (wft->refsta_seed, '_');
	if (ptr == NULL) {
		strcpy (wft->refsta, wft->refsta_seed);
		if (dbseed_net (wft->dbmaster, wft->refsta, snet, ssta) < 0) {
			register_error (0, "wftest_config: seed_net(refsta=%s) error.\n",
							wft->refsta_seed);
			return (-1);
		}
		sprintf (wft->refsta_seed, "%s_%s", snet, ssta);
	} else {
		*ptr = '\0';
		ptr++;
		strcpy (snet, wft->refsta_seed);
		strcpy (ssta, ptr);
		ptr--;
		*ptr = '_';
		dbmap_seed_netsta (wft->dbmaster, snet, ssta, wft->refsta);
		clear_register (1);
	}
	site = dblookup (wft->dbmaster, 0, "site", 0, 0);
	if (site.table == dbINVALID) {
		register_error (0, "wftest_config: Cannot lookup site table.\n");
		return (-1);
	}
	yd = yearday(wft->dbtime);
	sprintf (expr, "%d >= ondate && ( %d <= offdate || offdate == null(\"offdate\") ) && sta == \"%s\"", 
					yd, yd, wft->refsta);
	site = dbsubset (site, expr, NULL);
	dbquery (site, dbRECORD_COUNT, &n);
	if (n < 1) {
		register_error (0, "wftest_config: Cannot find refsta '%s' in site table for yearday %d.\n",
							wft->refsta, yd);
		return (-1);
	}
	if (n > 1) {
		register_error (0, "wftest_config: refsta '%s' appears more than once in site table for yearday %d.\n",
							wft->refsta, yd);
		return (-1);
	}
	site.record = 0;
	if (dbgetv (site, NULL, "lat", &(wft->latr), "lon", &(wft->lonr), NULL) < 0) {
		register_error (0, "wftest_config: dbgetv(lat,lon) error for refsta '%s'.\n",
							wft->refsta);
		return (-1);
	}

	if (parse_param (params, "slowness", P_DBL, 1, &(wft->slowness)) < 0) {
		register_error (0, "wftest_config: parse_param(slowness) error.\n");
		return (-1);
	}
	if (parse_param (params, "azimuth", P_DBL, 1, &(wft->azimuth)) < 0) {
		register_error (0, "wftest_config: parse_param(azimuth) error.\n");
		return (-1);
	}
	wft->xslowness = wft->slowness*sin(wft->azimuth*M_PI/180.0);
	wft->yslowness = wft->slowness*cos(wft->azimuth*M_PI/180.0);

	if (wftest_process_dbmaster(wftv) < 0) {
		register_error (0, "wftest_config: wftest_process_dbmaster() error.\n");
		return (-1);
	}

	wft->time = (double)((long)now()) + 1.0;
	wft->nexttime = wft->time + wft->nsamps / wft->samprate;
	if (!strcmp(wft->start, "m")) {
		int minutes;

		minutes = (int)((wft->nexttime + 60.5)/60.0);
		wft->tstart = (double)minutes*60.0;
		while (wft->tstart <= wft->nexttime) wft->tstart += 60.0;
	} else if (strcmp(wft->start, "")) {
		if (zstr2epoch (wft->start, &(wft->tstart)) < 0) {
			register_error (0, "wftest_config: zstr2epoch(%s) error for start time.\n", wft->start);
			return (-1);
		}
	}

	return (0);
}

int
wftest_proc (void *wftv, WFProcData *input, Tbl **output, Pf **state)

{
	WFTest *wft = (WFTest *) wftv;
	int i;
	double tstart, dt;

	if (input != NULL) {
		register_error (0, "wftest_proc: Wrong input type.\n");
		return (-1);
	}

	while (1) {
		double tnow;

		tnow = now();
		if (tnow > wft->nexttime) break;
		wfproc_usleep (0.1);
	}

	*output = newtbl (0);
	if (*output == NULL) {
		register_error (0, "wftest_proc: newtbl(output) error.\n");
		return (-1);
	}

	dt = 1.0/wft->samprate;
	for (i=0; i<maxtbl(wft->channels_tbl); i++) {
		WFTestChannel *wftc;
		WFProcData wfpd_tmp;
		size_t size;
		WFProcData *wfpd;
		int samp;
		double tdel, tstarts;

		wftc = (WFTestChannel *) gettbl (wft->channels_tbl, i);

		if (wftc->achan->data == NULL) {
			size = wft->nsamps*sizeof(int);
			wftc->achan->data = (int *) malloc (size);
			if (wftc->achan->data == NULL) {
				register_error (1, "wftest_proc: malloc(achan->data,%ld) error.\n", size);
				return (-1);
			}
			memset (wftc->achan->data, 0, size);
			wftc->achan->datasz = wft->nsamps;
		} else if (wft->nsamps > wftc->achan->datasz) {
			size = wft->nsamps*sizeof(int);
			wftc->achan->data = (int *) realloc (wftc->achan->data, size);
			if (wftc->achan->data == NULL) {
				register_error (1, "wftest_proc: realloc(achan->data,%ld) error.\n", size);
				return (-1);
			}
			memset (wftc->achan->data, 0, size);
			wftc->achan->datasz = wft->nsamps;
		}
	
		tstart = wft->tstart;
		tdel = wftc->delx*wft->xslowness + wftc->dely*wft->yslowness ;
		tstarts = tstart - tdel + wftc->tshift;
		if (!strcmp(wft->signal_type, "random")) {
			for (samp=0; samp<wft->nsamps; samp++) {
				double tvalue, dvalue;

				tvalue = wft->time + samp*dt;
				if (tstart > 0.0 && tvalue < tstarts) {
					dvalue = 0.0;
				} else if (tstart > 0.0 && wft->duration > 0.0 &&
							tvalue-tstarts > wft->duration) {
					dvalue = 0.0;
					if (wft->repeat > 0.0) {
						tstart += wft->repeat;
						tstarts = tstart - tdel + wftc->tshift;
					}
				} else {
					dvalue = disordered();
					dvalue -= 0.5;
					dvalue *= 2.0*wftc->amplitude;
				}
				wftc->achan->data[samp] = INT(dvalue);
			}
		} else if (!strcmp(wft->signal_type, "sine")) {
			for (samp=0; samp<wft->nsamps; samp++) {
				double tvalue, dvalue;

				tvalue = wft->time + samp*dt;
				if (tstart > 0.0 && tvalue < tstarts) {
					dvalue = 0.0;
				} else if (tstart > 0.0 && wft->duration > 0.0 &&
							tvalue-tstarts > wft->duration) {
					dvalue = 0.0;
					if (wft->repeat > 0.0) {
						tstart += wft->repeat;
						tstarts = tstart - tdel + wftc->tshift;
					}
				} else {
					dvalue = sin((tvalue-tstarts)*wft->param1);
					dvalue *= wftc->amplitude;
				}
				wftc->achan->data[samp] = INT(dvalue);
			}
		} else if (!strcmp(wft->signal_type, "dampedsine")) {
			for (samp=0; samp<wft->nsamps; samp++) {
				double tvalue, dvalue;

				tvalue = wft->time + samp*dt;
				if (tstart > 0.0 && tvalue < tstarts) {
					dvalue = 0.0;
				} else if (tstart > 0.0 && wft->duration > 0.0 &&
							tvalue-tstarts > wft->duration) {
					dvalue = 0.0;
					if (wft->repeat > 0.0) {
						tstart += wft->repeat;
						tstarts = tstart - tdel + wftc->tshift;
					}
				} else {
					dvalue = sin((tvalue-tstarts)*wft->param1);
					dvalue *= exp(-(tvalue-tstarts)*wft->param1*wft->param2);
					dvalue *= wftc->amplitude;
				}
				wftc->achan->data[samp] = INT(dvalue);
			}
		}
		wftc->achan->time = wft->time;
		wftc->achan->samprate = wft->samprate;
		wftc->achan->nsamp = wft->nsamps;

		size = sizeof(WFProcData);
		memset (&wfpd_tmp, 0, size);
		strcpy (wfpd_tmp.type, "wf");
		wfpd_tmp.data = wftc->achan;
		wfpd_tmp.printproc = wftest_data_print;
		wfpd_tmp.freeproc = wftest_data_free;
		wfpd_tmp.copyproc = wftest_data_copy;
		wfpd_tmp.topktproc = import_data_topkt;
		wfpd_tmp.size = size + sizeof(PktChannel) + 4*wftc->achan->datasz;
		wfpd = wftest_data_copy (&wfpd_tmp);
		if (wfpd == NULL) {
			register_error (0, "wftest_proc: wftest_data_copy(output) error.\n");
			return (-1);
		}
		settbl (*output, -1, wfpd);
	}

	wft->time = wft->nexttime;
	wft->nexttime = wft->time + wft->nsamps / wft->samprate;
	wft->tstart = tstart;

	return (0);
}

void
wftest_data_free (WFProcData *packet)

{
	PktChannel *chan;

	if (packet == NULL) return;
	chan = (PktChannel *) packet->data;
	if (chan != NULL) {
		freePktChannel (chan);
	}
	free (packet);
}

char *
wftest_data_print (WFProcData *packet, int verbose)

{
	PktChannel *chan;
	char line[512];
	char *s;

	if (packet == NULL) return (NULL);
	chan = (PktChannel *) packet->data;
	if (chan == NULL) return (NULL);
	s = strtime(chan->time);
	if (!strcmp(chan->loc, "") || !strcmp(chan->loc, "-")) {
		sprintf (line, "%s_%s_%s %s %d %.3f", chan->net, chan->sta, chan->chan,
				s, chan->nsamp, chan->samprate);
	} else {
		sprintf (line, "%s_%s_%s_%s %s %d %.3f", chan->net, chan->sta, chan->chan,
				chan->loc, s, chan->nsamp, chan->samprate);
	}
	free (s);

	return (strdup(line));
}

WFProcData *
wftest_data_copy (WFProcData *packet)

{
	WFProcData *copy;
	size_t size;
	PktChannel *chan, *chancopy;

	if (packet == NULL) return (NULL);

	size = sizeof(WFProcData);
	copy = (WFProcData *) malloc (size);
	if (copy == NULL) {
		register_error (1, "wftest_data_copy: malloc(WFProcData,%ld) error.\n", size);
		return (NULL);
	}

	chancopy = newPktChannel();
	if (chancopy == NULL) {
		free (copy);
		register_error (0, "wftest_data_copy: newPktChannel() error.\n");
		return (NULL);
	}

	chan = (PktChannel *)packet->data;
	*chancopy = *chan;
	chancopy->datasz = chancopy->nsamp;
	size = chancopy->nsamp*sizeof(int);
	chancopy->data = (int *) malloc (size);
	if (chancopy->data == NULL) {
		free (copy);
		free (chan);
		register_error (1, "wftest_data_copy: malloc(chancopy->data,%ld) error.\n", size);
		return (NULL);
	}

	memcpy (chancopy->data, chan->data, size);

	*copy = *packet;
	copy->data = chancopy;

	return (copy);
}
