/* Copyright (c) 2008 Boulder Real Time Technologies, Inc. */
/* All rights reserved */
 
/* This software module is wholly owned by Boulder Real Time 
   Technologies, Inc. This software may be used freely in any 
   way as long as the copyright statement above is not removed. */

#ifndef _wfgather_
#define _wfgather_

#include "stock.h"
#include "wfproc.h"

typedef struct wfgather_channel_ {
	char snet[32];
	char ssta[32];
	char schan[32];
	char sloc[32];
	char sta[32];
	char chan[32];
	char segtype[32];
	char filter[512];
	double samprate;
	double calib;
	double calper;
	double lat;
	double lon;
	double elev;
	double delx;
	double dely;
	double clip_lower;
	double clip_upper;
	int ichannel;
	Hook *filter_hook;
} WFGatherChannel;

typedef struct wfgather_ {
	WFProcDb *wfpdbm;
	Dbptr dbmaster;
	double time;
	Tbl *channels_accept;
	Tbl *channels_reject;
	Arr *channels_arr;
	Tbl *channels_tbl;
	char start[32];
	double output_duration;
	double maximum_wait_time;
	double tstart;
	double samprate;
	int verbose;
	int nsamps_per_tblock;
	int nvalues_per_tblock;
	int ntblocks;
	int nchannels;
	int nvalues;
	int nsamps;
	int block_wall_times_nentries;
	double *block_wall_times;
	float clip_value;
	float gap_value;
	float *gather;
} WFGather;

typedef struct wfgatheroutput_ {
	Tbl *channels;
	double tstart;
	double samprate;
	int nchannels;
	int nsamps;
	int nvalues;
	int ngaps;
	int nclips;
	float clip_value;
	float gap_value;
	float *gather;
} WFGatherOutput;

int wfgather_make_nscl (char *net, char *sta, char *chan, char *loc, char *nscl);
int wfgather_parse_nscl (char *nscl, char *net, char *sta, char *chan, char *loc);
int wfgather_match_channel (void *wfgv, char *nscl);
int wfgather_process_dbmaster (void *wfgv);
void *wfgather_new (Pf *params);
int wfgather_destroy (void *wfgv);
int wfgather_config (void *wfgv, Pf *params, Pf *state);
int wfgather_proc (void *wfgv, WFProcData *input, Tbl **output, Pf **state);
WFProcData *wfgather_create_output (WFGather *wfg, int iblock);
char *wfgather_data_print (WFProcData *data, int verbose);
void wfgather_data_free (WFProcData *data);
WFProcData *wfgather_data_copy (WFProcData *data);
int wfgather_data_topkt (WFProcData *data, Tbl *translations, Tbl *pkts_tbl);
Tbl *wfgather_channels_copy (Tbl *channels);
int wfgather_channels_match (Tbl *channels1, Tbl *channels2);

#endif
